BONUSSCAN_PATTERN_SETNAME = "^(.*) %(%d/%d%)$";

BONUSSCAN_PATTERN_GENERIC_PREFIX = "^%+(%d+)%%?(.*)$";
BONUSSCAN_PATTERN_GENERIC_SUFFIX = "^(.*)%+(%d+)%%?$";

BONUSSCAN_VERSION = "v0.9";

BonusScan_bonuses = {};
BonusScan_bonuses_details = {};

BonusScan_active = nil;
BonusScan_temp = { 
	sets = {},
	set = "",
	slot = "",
	bonuses = {},
	details = {}
}

BonusScan_types = {
	"STR", 		-- strength	
	"AGI", 		-- agility
	"STA", 		-- stamina
	"INT", 		-- intellect
	"SPI", 		-- spirit
	"ARMOR", 	-- reinforced armor (not base armor)
  
	"ARCANERES",-- arcane resistance
	"FIRERES",  -- fire resistance
	"NATURERES",-- nature resistance 	
	"FROSTRES", -- frost resistance
	"SHADOWRES",-- shadow resistance

	"FISHING",  -- fishing skill
	"MINING",	-- mining skill
	"HERBALISM",-- herbalism skill
	"SKINNING", -- skinning skill
	"DEFENSE",  -- defense skill

	"BLOCK",    -- chance to block
	"DODGE",	-- chance to dodge
	"PARRY",	-- chance to parry
	"ATTACKPOWER", -- attack power
	"CRIT",		-- chance to get a critical strike
	"RANGEDATTACKPOWER", -- ranged attack power
	"RANGEDCRIT", -- chance to get a crit with ranged weapons
	"TOHIT",	-- chance to hit

	"DMG",		-- spell damage
	"ARCANEDMG",-- arcane spell damage
	"FIREDMG",	-- fire spell damage
	"FROSTDMG",	-- frost spell damage
	"HOLYDMG",	-- holy spell damage
	"NATUREDMG",-- nature spell damage
	"SHADOWDMG",-- shadow spell damage
	"SPELLCRIT",-- chance to crit with spells
	"HEAL",		-- healing 
	"HOLYCRIT", -- chance to crit with holy spells

	"HEALTHREG",-- health regeneration per 5 sec.
	"MANAREG",	-- mana regeneration per 5 sec.
	"HEALTH",	-- health points
	"MANA",		-- mana points
}

BonusScan_slots = {
	"Head",
	"Neck",
	"Shoulder",
	"Shirt",
	"Chest",
	"Waist",
	"Legs",
	"Feet",
	"Wrist",
	"Hands",
	"Finger0",
	"Finger1",
	"Trinket0",
	"Trinket1",
	"Back",
	"MainHand",
	"SecondaryHand",
	"Ranged",
	"Tabard",
};


-- Update function to hook into. 
-- Gets called, when Equipment changes (after UNIT_INVENTORY_CHANGED)
function BonusScan_Update()
end

function BonusScan_GetBonus(bonus)
	if(BonusScan_bonuses[bonus]) then
		return BonusScan_bonuses[bonus];
	end;
	return 0;
end

function BonusScan_GetSlotBonuses(slotname)
	local i, bonus, details;
	local bonuses = {};
	for bonus, details in BonusScan_bonuses_details do
		if(details[slotname]) then
			bonuses[bonus] = details[slotname];
		end
	end
	return bonuses;
end

function BonusScan_GetBonusDetails(bonus)
	if(BonusScan_bonuses_details[bonus]) then
		return BonusScan_bonuses_details[bonus];
	end;
	return {};
end

function BonusScan_GetSlotBonus(bonus, slotname)
	if(BonusScan_bonuses_details[bonus]) then
		if(BonusScan_bonuses_details[bonus][slotname]) then
			return BonusScan_bonuses_details[bonus][slotname];
		end;
	end;
	return 0;
end


function BonusScan_OnLoad()
	this:RegisterEvent("PLAYER_LOGIN");
	this:RegisterEvent("UNIT_INVENTORY_CHANGED");
end

function BonusScan_OnEvent()
	if (event == "PLAYER_LOGIN") then
		BonusScan_active = 1;
	end
	if (((event == "PLAYER_LOGIN") or (event == "UNIT_INVENTORY_CHANGED")) and BonusScan_active) then
		BonusScan_Time = time();
		if (BonusScan_Time ~= BonusScan_LastTime) then
			BonusScan_ScanEquipment();
			BonusScan_Update();
			BonusScan_LastTime = BonusScan_Time;
		end
	end
end


function BonusScan_ScanEquipment()
	local slotid, slotname, hasItem, i;

	BonusScan_temp.bonuses = {};
	BonusScan_temp.details = {};
	BonusScan_temp.sets = {};
	BonusScan_temp.set = "";

	for i, slotname in BonusScan_slots do
		slotid, _ = GetInventorySlotInfo(slotname.. "Slot");
		hasItem = BonusScanTooltip:SetInventoryItem("player", slotid);
	
		if ( hasItem ) then
			BonusScan_temp.slot = slotname;
			BonusScan_ScanTooltip();
			-- if set item, mark set as already scanned
			if(BonusScan_temp.set ~= "") then
				BonusScan_temp.sets[BonusScan_temp.set] = 1;
			end;
		end
	end
	BonusScan_bonuses = BonusScan_temp.bonuses;
	BonusScan_bonuses_details = BonusScan_temp.details;
end

function BonusScan_ScanItem(itemlink)
	local name = GetItemInfo(itemlink);
	if(name) then
		BonusScan_temp.bonuses = {};
		BonusScan_temp.sets = {};
		BonusScan_temp.set = "";
		BonusScan_temp.slot = "";
		BonusScanTooltip:SetHyperlink(itemlink);
		BonusScan_ScanTooltip();
		return BonusScan_temp.bonuses;
	end
	return false;
end

function BonusScan_ScanTooltip()
	local tmpTxt, line;
	local lines = BonusScanTooltip:NumLines();

	for i=2, lines, 1 do
		tmpText = getglobal("BonusScanTooltipTextLeft"..i);
		val = nil;
		if (tmpText:GetText()) then
			line = tmpText:GetText();
			BonusScan_ScanLine(line);
		end
	end
end

function BonusScan_AddValue(effect, value)
	local i,e;
	
	if(type(effect) == "string") then
		if(BonusScan_temp.bonuses[effect]) then
			BonusScan_temp.bonuses[effect] = BonusScan_temp.bonuses[effect] + value;
		else
			BonusScan_temp.bonuses[effect] = value;
		end
		
		if(BonusScan_temp.slot) then
			if(BonusScan_temp.details[effect]) then
				if(BonusScan_temp.details[effect][BonusScan_temp.slot]) then
					BonusScan_temp.details[effect][BonusScan_temp.slot] = BonusScan_temp.details[effect][BonusScan_temp.slot] + value;
				else
					BonusScan_temp.details[effect][BonusScan_temp.slot] = value;
				end
			else
				BonusScan_temp.details[effect] = {};
				BonusScan_temp.details[effect][BonusScan_temp.slot] = value;
			end
		end;
	else 
	-- list of effects
		for i,e in effect do
			BonusScan_AddValue(e, value);
		end
	end
end;

function BonusScan_ScanLine(line)
	local tmpStr, found;
	
	-- Check for "Equip: "
	if(string.sub(line,0,string.len(BONUSSCAN_PREFIX_EQUIP)) == BONUSSCAN_PREFIX_EQUIP) then

		tmpStr = string.sub(line,string.len(BONUSSCAN_PREFIX_EQUIP)+1);
		BonusScan_CheckPassive(tmpStr);

	-- Check for "Set: "
	elseif(string.sub(line,0,string.len(BONUSSCAN_PREFIX_SET)) == BONUSSCAN_PREFIX_SET
			and BonusScan_temp.set ~= "" 
			and not BonusScan_temp.sets[BonusScan_temp.set]) then

		tmpStr = string.sub(line,string.len(BONUSSCAN_PREFIX_SET)+1);
		BonusScan_temp.slot = "Set";
		BonusScan_CheckPassive(tmpStr);

	-- any other line (standard stats, enchantment, set name, etc.)
	else
		-- Check for set name
		_, _, tmpStr = string.find(line, BONUSSCAN_PATTERN_SETNAME);
		if(tmpStr) then
			BonusScan_temp.set = tmpStr;
		else
			found = BonusScan_CheckGeneric(line);
			if(not found) then
				BonusScan_CheckOther(line);
			end;
		end
	end
end;


-- Scans passive bonuses like "Set: " and "Equip: "
function BonusScan_CheckPassive(line)
	local i, p, value, found;

	found = nil;
	for i,p in BONUSSCAN_PATTERNS_PASSIVE do
		_, _, value = string.find(line, "^" .. p.pattern);
		if(value) then
			BonusScan_AddValue(p.effect, value)
			found = 1;
		end
	end
	if(not found) then
		BonusScan_CheckGeneric(line);
	end
end


-- Scans generic bonuses like "+3 Intellect" or "Arcane Resistance +4"
function BonusScan_CheckGeneric(line)
	local value, token, pos, tmpStr, found;

	-- split line at "/" for enchants with multiple effects
	found = false;
	while(string.len(line) > 0) do
		pos = string.find(line, "/", 1, true);
		if(pos) then
			tmpStr = string.sub(line,1,pos-1);
			line = string.sub(line,pos+1);
		else
			tmpStr = line;
			line = "";
		end

		-- trim line
	    tmpStr = string.gsub( tmpStr, "^%s+", "" );
   	    tmpStr = string.gsub( tmpStr, "%s+$", "" );
       	tmpStr = string.gsub( tmpStr, "%.$", "" );

		_, _, value, token = string.find(tmpStr, BONUSSCAN_PATTERN_GENERIC_PREFIX);
		if(not value) then
			_, _,  token, value = string.find(tmpStr, BONUSSCAN_PATTERN_GENERIC_SUFFIX);
		end
		if(token and value) then
			-- trim token
		    token = string.gsub( token, "^%s+", "" );
    	    token = string.gsub( token, "%s+$", "" );
	       	token = string.gsub( token, "%.$", "" );
	
			if(BonusScan_CheckToken(token,value)) then
				found = true;
			end
		end
	end
	return found;
end


-- Identifies simple tokens like "Intellect" and composite tokens like "Fire damage" and 
-- add the value to the respective bonus. 
-- returns true if some bonus is found
function BonusScan_CheckToken(token, value)
	local i, p, s1, s2;
	
	if(BONUSSCAN_PATTERNS_GENERIC_LOOKUP[token]) then
		BonusScan_AddValue(BONUSSCAN_PATTERNS_GENERIC_LOOKUP[token], value);
		return true;
	else
		s1 = nil;
		s2 = nil;
		for i,p in BONUSSCAN_PATTERNS_GENERIC_STAGE1 do
			if(string.find(token,p.pattern,1,1)) then
				s1 = p.effect;
			end
		end	
		for i,p in BONUSSCAN_PATTERNS_GENERIC_STAGE2 do
			if(string.find(token,p.pattern,1,1)) then
				s2 = p.effect;
			end
		end	
		if(s1 and s2) then
			BonusScan_AddValue(s1..s2, value);
			return true;
		end 
	end
	return false;
end

-- Last fallback for not generic enchants, like "Mana Regen x per 5 sec."
function BonusScan_CheckOther(line)
	local i, p, value, start, found;

	for i,p in BONUSSCAN_PATTERNS_OTHER do
		start, _, value = string.find(line, "^" .. p.pattern);
		if(start) then
			if(p.value) then
				BonusScan_AddValue(p.effect, p.value)
			elseif(value) then
				BonusScan_AddValue(p.effect, value)
			end
			return true;
		end
	end
	return false;
end



-- Slash Command functions

function BonusScan_Cmd(cmd)

	local _, _, itemlink, itemid = string.find(cmd, "|c%x+|H(item:(%d+):%d+:%d+:%d+)|h%[.-%]|h|r");
	if(itemid) then
  		local name = GetItemInfo(itemid);
		if(name) then
			DEFAULT_CHAT_FRAME:AddMessage(GREEN_FONT_COLOR_CODE .. "Item bonuses of: " .. HIGHLIGHT_FONT_COLOR_CODE .. name);
			local bonuses = BonusScan_ScanItem(itemlink);
			if(not bonuses) then
				DEFAULT_CHAT_FRAME:AddMessage(GREEN_FONT_COLOR_CODE .. "error scanning item (probably not cached)");
			else
	  			BonusScan_PrintInfo(bonuses);
			end
  		end
  		return;
  	end
  	if(string.lower(cmd) == "show") then
	  	DEFAULT_CHAT_FRAME:AddMessage(GREEN_FONT_COLOR_CODE .. "Current equipment bonuses:");
		BonusScan_PrintInfo(BonusScan_bonuses);
  		return;
  	end
  	if(string.lower(cmd) == "details") then
	  	DEFAULT_CHAT_FRAME:AddMessage(GREEN_FONT_COLOR_CODE .. "Current equipment bonus details:");
		BonusScan_PrintInfoDetailed();
  		return;
  	end
  	DEFAULT_CHAT_FRAME:AddMessage(GREEN_FONT_COLOR_CODE .. "BonusScan " .. BONUSSCAN_VERSION .. " by Crowley");
  	DEFAULT_CHAT_FRAME:AddMessage(GREEN_FONT_COLOR_CODE .. "/bscan show - shows all bonus of the current equipment");
  	DEFAULT_CHAT_FRAME:AddMessage(GREEN_FONT_COLOR_CODE .. "/bscan details - shows bonuses with slot distribution");
  	DEFAULT_CHAT_FRAME:AddMessage(GREEN_FONT_COLOR_CODE .. "/bscan <itemlink> - shows bonuses of linked item (insert link with Shift-Click)");
end

SLASH_BONUSSCAN1 = "/bonusscanner";
SLASH_BONUSSCAN2 = "/bscan";
SlashCmdList["BONUSSCAN"] = BonusScan_Cmd;


function BonusScan_PrintInfoDetailed()
	local bonus, name, i, j, slot, first, s;
	for i, bonus in BonusScan_types do
		if(BonusScan_bonuses[bonus]) then
			first = true;
			s = "(";
			for j, slot in BonusScan_slots do 
				if(BonusScan_bonuses_details[bonus][slot]) then
					if(not first) then
						s = s .. ", ";
					else
						first = false;
					end
					s = s .. 	LIGHTYELLOW_FONT_COLOR_CODE .. slot .. 
								HIGHLIGHT_FONT_COLOR_CODE .. ": " .. BonusScan_bonuses_details[bonus][slot];
				end
			end;
			if(BonusScan_bonuses_details[bonus]["Set"]) then
				if(not first) then
					s = s .. ", ";
				end
				s = s .. 	LIGHTYELLOW_FONT_COLOR_CODE .. "Set" .. 
							HIGHLIGHT_FONT_COLOR_CODE .. ": " .. BonusScan_bonuses_details[bonus]["Set"];
				end
			s = s .. ")";
			DEFAULT_CHAT_FRAME:AddMessage(GREEN_FONT_COLOR_CODE .. BONUSSCAN_NAMES[bonus] .. ": " .. HIGHLIGHT_FONT_COLOR_CODE .. BonusScan_bonuses[bonus] .. " " .. s);
		end
	end
end

function BonusScan_PrintInfo(bonuses)
	local bonus, i;
	for i, bonus in BonusScan_types do
		if(bonuses[bonus]) then
			DEFAULT_CHAT_FRAME:AddMessage(GREEN_FONT_COLOR_CODE .. BONUSSCAN_NAMES[bonus] .. ": " .. HIGHLIGHT_FONT_COLOR_CODE .. bonuses[bonus]);
		end
	end
end
