--[[

    ColorCycle

    Written by Cragganmore

    This mod allows you to change the colors of various UI elements to create
    effects such as glows, fades, or color-cycles.
    
  ]]--

-- Configuration Constants
COLORCYCLE_VERSION     = "1.00";
COLORCYCLE_DEBUG       = false;
COLORCYCLE_TRON        = false;
COLORCYCLE_DEBUG_LEVEL = 10;

-- Global Constants
COLORCYCLE_DEFAULT_CYCLESPEED           = 0.05;
COLORCYCLE_DEFAULT_VARYSPEEDFACTOR      = 0;
COLORCYCLE_DEFAULT_STARTVALUES          = {false, false, false};
COLORCYCLE_DEFAULT_ENDVALUES            = {false, false, false};
COLORCYCLE_DEFAULT_LOCKEDVALUES         = {false, false, false};
COLORCYCLE_DEFAULT_REDBOUNDS            = {0.0, 1.0};
COLORCYCLE_DEFAULT_GREENBOUNDS          = {0.0, 1.0};
COLORCYCLE_DEFAULT_BLUEBOUNDS           = {0.0, 1.0};
COLORCYCLE_DEFAULT_FOLLOWLIST           = {0, 0, 0};
COLORCYCLE_DEFAULT_OKTOWAIT             = {true, true, true};
COLORCYCLE_DEFAULT_PAUSEATLOWERBOUNDS   = {0, 0, 0};
COLORCYCLE_DEFAULT_PAUSEATUPPERBOUNDS   = {0, 0, 0};
COLORCYCLE_DEFAULT_PAUSE                = 1;
COLORCYCLE_DEFAULT_GLOWSPEED            = 0.05;
COLORCYCLE_CYCLETYPES                   = {"ToolTipText", "FontText", "Texture", 
                                           "FrameBackground", "FrameBorder", "StatusBar"};

-- used to change the color of text in the chat box
COLORCYCLE_COLOR_RED       = "|cffff0000";
COLORCYCLE_COLOR_YELLOW    = "|cffffff00";
COLORCYCLE_COLOR_VIOLET    = "|cffbb88dd";
COLORCYCLE_COLOR_GREY      = "|caaaaaaaa";
COLORCYCLE_COLOR_WHITE     = "|cffffffff";
COLORCYCLE_COLOR_BLUE      = "|cff3366ff";
COLORCYCLE_COLOR_GREEN     = "|cff64c528";
COLORCYCLE_COLOR_LTGREEN   = "|cffE4FB75";
COLORCYCLE_COLOR_CLOSE     = "|r";

-- Configuration Variables (saved across sessions)
ColorCycle_Config = { };
ColorCycle_Config.Enabled = 1;
ColorCycle_Config.Version = COLORCYCLE_VERSION;
function ColorCycle_ResetConfig()
    ColorCycle_Config.Enabled = true;
    ColorCycle_Config.Version = COLORCYCLE_VERSION;
end

-- Variables
Cycle_Table = { };

COLORCYCLE_BASECOLORS = { ["white"]             = {1.00, 1.00, 1.00},
                          ["black"]             = {0.00, 0.00, 0.00},    

                          ["red"]               = {1.00, 0.00, 0.00},    
                          ["lightred"]          = {1.00, 0.50, 0.50},    
                          ["orange"]            = {1.00, 0.50, 0.00},    
                          ["lightorange"]       = {1.00, 0.75, 0.50},    
                          ["gold"]              = {1.00, 0.75, 0.00},    
                          ["lightgold"]         = {1.00, 0.87, 0.50},    

                          ["yellow"]            = {1.00, 1.00, 0.00},    
                          ["lightyellow"]       = {1.00, 1.00, 0.50},    
                          ["yellowgreen"]       = {0.50, 1.00, 0.00},    
                          ["lightyellowgreen"]  = {0.75, 1.00, 0.50},    
                          ["green"]             = {0.00, 1.00, 0.00},    
                          ["lightgreen"]        = {0.50, 1.00, 0.50},    

                          ["bluegreen"]         = {0.00, 1.00, 0.50},    
                          ["lightbluegreen"]    = {0.50, 1.00, 0.75},    
                          ["cyan"]              = {0.00, 1.00, 1.00},    
                          ["lightcyan"]         = {0.50, 1.00, 1.00},    
                          ["blue"]              = {0.00, 0.00, 1.00},    
                          ["lightblue"]         = {0.50, 0.50, 1.00},    

                          ["bluepurple"]        = {0.50, 0.00, 1.00},    
                          ["lightbluepurple"]   = {0.75, 0.50, 1.00},    
                          ["purple"]            = {1.00, 0.00, 1.00},    
                          ["lightpurple"]       = {1.00, 0.50, 1.00},    
                          ["pinkred"]           = {1.00, 0.00, 0.50},    
                          ["lightpinkred"]      = {1.00, 0.50, 0.75}     
                        };


--============================================================================================--
--============================================================================================--
--                                                                                            --
--                              LOADING AND INIT FUNCTIONS                                    --
--                                                                                            --
--============================================================================================--
--============================================================================================--



-----------------------------------------------------------------------------------
-- Handle the loading stuff when the plugin first loads
-----------------------------------------------------------------------------------
function ColorCycle_OnLoad()       
    SlashCmdList["COLORCYCLE_SLASHENABLE"] = ColorCycle_SlashCommandHandler;
    SLASH_COLORCYCLE_SLASHENABLE1 = "/colorcycle";
    SLASH_COLORCYCLE_SLASHENABLE2 = "/cc";

    -- make sure we have the Print() function
    if (not Print) then
        function Print(msg, r, g, b, frame) 
            if (not r) then r = 1.0; end
            if (not g) then g = 1.0; end
            if (not b) then b = 1.0; end
            if ( frame ) then 
                frame:AddMessage(msg,r,g,b);
            else
                if ( DEFAULT_CHAT_FRAME ) then 
                    DEFAULT_CHAT_FRAME:AddMessage(msg, r, g, b);
                end
            end
        end
    end
    
end


--============================================================================================--
--============================================================================================--
--                                                                                            --
--                              EVENT AND SLASH COMMAND FUNCTIONS                             --
--                                                                                            --
--============================================================================================--
--============================================================================================--


---------------------------------------------------------------------------------
-- Break a chat command into its command and variable parts (i.e. "debug on"
-- breaks into command = "debug" and variable = "on"
---------------------------------------------------------------------------------
function ColorCycle_ParseCommand(msg)
    firstSpace = string.find(msg, " ", 1, true);
    if (firstSpace) then
        local command = string.sub(msg, 1, firstSpace - 1);
        local var  = string.sub(msg, firstSpace + 1);
        return command, var
    else
        return msg, nil;
    end
end



---------------------------------------------------------------------------------
-- A simple chat command handler.   takes commands in the form "/generic command var"
---------------------------------------------------------------------------------
function ColorCycle_SlashCommandHandler(msg)

    local command, var = ColorCycle_ParseCommand(msg);
    if ((not command) and msg) then
        command = msg;
    end
    if (command) then
        command = string.lower(command);
        if (command == COLORCYCLE_VERSION_CMD) then
            Print("ColorCycle Version "..COLORCYCLE_VERSION);
        elseif (command == COLORCYCLE_REPORT_CMD) then
            ColorCycle_Report(false);
        elseif (command == COLORCYCLE_REPORTFULL_CMD) then
            ColorCycle_Report(true, var);
        elseif (command == "q") then
            ColorCycle_QList();
        elseif (command == COLORCYCLE_DEBUG_CMD) then
            ColorCycle_ToggleDebug();
        elseif (command == COLORCYCLE_ENABLE_CMD or command == COLORCYCLE_ON_CMD) then
            ColorCycle_Toggle(1);
        elseif (command == COLORCYCLE_DISABLE_CMD or command == COLORCYCLE_OFF_CMD) then
            ColorCycle_Toggle(0);
        elseif (command == "examples") then
            if (ColorCycle_ExamplifyIt) then 
                ColorCycle_ExamplifyIt(var); 
            else
                Print("ColorCycle examples couldn't be found.");
            end
        elseif (command == COLORCYCLE_HELP_CMD or command == "") then
            ColorCycle_ShowHelp();
        else
            ColorCycle_ShowHelp();        
        end
    end
end

    

--============================================================================================--
--============================================================================================--
--                                                                                            --
--                              MAIN FUNCTIONS                                                --
--                                                                                            --
--============================================================================================--
--============================================================================================--

-----------------------------------------------------------------------------------
-- Add an entry to the list of stuff to color cycle
--
-- required parameters:
-- -------------------
-- ID:              a string the uniquely identifies this color cycle
-- cycleType:       can be "ToolTipText", "FontText", "Texture", "FrameBackground", "FrameBorder", "StatusBar"
-- globalName:      name of the global variable associated with the object to color cycle (ie variable for getglobal())
--
-- optional parameters:
-- -------------------
-- cycleSpeed:          the amount to increase or decrease each color channel when onUpdate gets called (from 
--                      0.0 to 1.0, defaults to 0.05).
-- varySpeedFactor:     if you want the cycle speed to vary once in a while, set this to a value (between 0.5 and 2 
--                      are reasonable).  if not, set to 0.
-- colorStart:         {r, g, b} table that specifies the min value for each channel (from 0.0 to 1.0).
-- colorEnd:           {r, g, b} table that specifies the max value for each channel (from 0.0 to 1.0).
-- lockedValues:        {r, g, b} table.  if you want to lock the red channel to a certain value (1.0, 0.0, etc), 
--                      set the first element of this table to the value.  this would always keep red at 1.0.  
--                      if you don't want to lock a channel, set it to nil.
-- redBounds:           {min, max} table.  to set the bounds for the red channel between 0.2 and 0.8, this would
--                      be {0.2, 0.8}.  to use default bounds of {0.0, 1.0} you can leave this nil.
-- greenBounds:         {min, max} table.  to set the bounds for the green channel between 0.2 and 0.8, this would
--                      be {0.2, 0.8}.  to use default bounds of {0.0, 1.0} you can leave this nil.
-- blueBounds:          {min, max} table.  to set the bounds for the blue channel between 0.2 and 0.8, this would
--                      be {0.2, 0.8}.  to use default bounds of {0.0, 1.0} you can leave this nil.
-- followList:          {r, g, b} table that specifies which (if any) channels should stay in sync with each other.
--                      for instance, to make the red and blue channels always have the same values, set this to
--                      {1, 0, 1}.  to keep all the channels independent, set this to nil.
-- okToWait:            {boolean, boolean, boolean} table that specifies if it's ok for each channel to randomly 
--                      stop moving and wait for another channel to reach a specified value.  Creates more random 
--                      cycling effects.
-- pauseAtLowerBounds:  {secs, secs, secs} table that specifies the number of seconds a channel should pause (not change)
--                      when it reaches its lower bounds.  this can be used to create pulsing effects.
-- pauseAtUpperBounds:  {secs, secs, secs} table that specifies the number of seconds a channel should pause (not change)
--                      when it reaches its upper bounds.  this can be used to create pulsing effects.
-- text:                the text value, if any, for the object to be cycled (like if you need to cycle text in a tooltip, 
--                      you'll have to provide the text since you can't set tooltip text color without also providing 
--                      the text).
-- stayBright:          set this to true if you never want all three color channels to ever be 0 (black)
-- startingFunction:    a function that will be called when this cycle starts
-- endingFunction:      a function that will be called when this cycle ends
-- endCondition:        {r, g, b} table.  when the colors cycle to these values, this cycle will end
-- goto:                what to do when we end this cycle: "end" = stop this cycle, "random" = pick a random entry in the queue
--                      to start cycling at, a number = go to this entry in the queue, nil = move to the next entry in the queue
--                      (looping back to the first entry)
-- unique:              set to true if you don't want to allow more than one entry for this ID
--
--
-- ColorCycle_AddTableEntry(ID                 = "string";
--                          cycleType          = "ToolTipText" | "FontText" | "Texture" | "FrameBackground" | "FrameBorder" | "StatusBar";
--                          globalName         = "string"; 
--                          cycleSpeed         = number;
--                          varySpeedFactor    = number;
--                          colorStart         = {startRedValue, startGreenValue, startBlueValue};
--                          colorEnd           = {endRedValue, endGreenValue, endBlueValue};
--                          lockedValues       = {lockValueR, lockValueG, lockValueB};
--                          redBounds          = {minR, maxR};
--                          greenBounds        = {minG, maxG};
--                          blueBounds         = {minB, maxB};
--                          followList         = {followR, followG, followB};
--                          okToWait           = {OKr, OKg, OKb};
--                          pauseAtLowerBounds = {redSecs, greenSecs, blueSecs};
--                          pauseAtUpperBounds = {redSecs, greenSecs, blueSecs};
--                          text               = "string";
--                          stayBright         = true | false;
--                          startingFunction   = functionName;
--                          endingFunction     = functionName;
--                          endCondition       = {r, g, b};
--                          goto               = "end" | "random" | number | nil;
--                          unique             = true | false;                        
--                          )
--
-----------------------------------------------------------------------------------
function ColorCycle_AddTableEntry(...)
    ColorCycleTron("ColorCycle_AddTableEntry");
    
    local noproblems = true;

    ColorCycleDebug("AddTableEntry called with "..table.getn(arg).." entries to add.");
    
    -- for each entry passed in
    for i = 1, table.getn(arg) do      
        local entry = arg[i];
        local valid = false;
        if (entry.unique == nil) then entry.unique = true; end
        
        -- if we're not already cycling this UI element
        if (not entry.unique or (entry.unique and (not ColorCycle_InCycleTable(entry.ID)))) then
            ColorCycleDebug("AddTableEntry: entry = "..asText(entry));
            valid, entry = ColorCycle_ValidateTableEntry(entry);
            ColorCycleDebug("AddTableEntry: after ColorCycle_ValidateTableEntry, entry = "..asText(entry));
            if (valid) then
                ColorCycle_ProcessNewTableEntry(entry);
            else
                noproblems = false;
            end
        end    
    end
    
    return noproblems;
end

-----------------------------------------------------------------------------------
-- Make sure the right table elements were passed in
-- All this function does is error-check that the
-- right parameters were passed to ColorCycle_AddTableEntry()
-----------------------------------------------------------------------------------
function ColorCycle_ValidateTableEntry(v)
    local valid = true;
    
    if (not v.ID) then
        ColorCyclePrint("No ID passed to ColorCycle_AddTableEntry!");
        return false;
    end

    if (not v.globalName) then
        ColorCyclePrint("No globalName passed to ColorCycle_AddTableEntry!");
        return false;
    end

    if (not v.cycleType) then 
        ColorCyclePrint("No cycleType passed to ColorCycle_AddTableEntry for UI element "..v.globalName.."!");
        return false;
    end

    if (not v.startingFunction) then 
        v.startingFunction = nil;
    end

    if (not v.endingFunction) then 
        v.endingFunction = nil;
    end

    if (not v.endCondition) then 
        v.endCondition = nil;
    end

    if (not v.goto) then 
        v.goto = nil;
    end
    
    if (not v.stayBright) then
        v.stayBright = false;
    end    

    if (not v.started) then 
        v.started = false;
    end

    if (not v.cycleSpeed or v.cycleSpeed <= 0) then 
        v.cycleSpeed = COLORCYCLE_DEFAULT_CYCLESPEED;
    end

    if (not v.varySpeedFactor) then 
        v.varySpeedFactor = COLORCYCLE_DEFAULT_VARYSPEEDFACTOR;
    end

    if (not v.colorStart or v.colorStart == {}) then 
        v.colorStart = ColorCycle_DefaultStartValues();
    end

    if (not v.colorEnd or v.colorEnd == {}) then 
        v.colorEnd = ColorCycle_DefaultEndValues();
    end

    if (not v.lockedValues or v.lockedValues == {}) then 
        v.lockedValues = ColorCycle_DefaultLockedValues();
    end

    if (not v.redBounds or v.redBounds == {}) then 
        v.redBounds = ColorCycle_DefaultRedBounds();
    end

    if (not v.greenBounds or v.greenBounds == {}) then 
        v.greenBounds = ColorCycle_DefaultGreenBounds();
    end

    ColorCycleDebug("ValidateTableEntry before check v.blueBounds = "..asText(v.blueBounds));
    ColorCycleDebug("ValidateTableEntry before check COLORCYCLE_DEFAULT_BLUEBOUNDS = "..asText(COLORCYCLE_DEFAULT_BLUEBOUNDS));
    if (not v.blueBounds or v.blueBounds == {}) then 
        v.blueBounds = ColorCycle_DefaultBlueBounds();
    end
    ColorCycleDebug("ValidateTableEntry after check v.blueBounds = "..asText(v.blueBounds));
    ColorCycleDebug("ValidateTableEntry after check COLORCYCLE_DEFAULT_BLUEBOUNDS = "..asText(COLORCYCLE_DEFAULT_BLUEBOUNDS));

    if (not v.followList or v.followList == {}) then 
        v.followList = ColorCycle_DefaultFollowList();
    end

    if (not v.okToWait or v.okToWait == {}) then 
        v.okToWait = ColorCycle_DefaultOkToWait();
    end

    if (not v.pauseAtLowerBounds or v.pauseAtLowerBounds == {}) then 
        v.pauseAtLowerBounds = ColorCycle_DefaultPauseAtLowerBounds();
    end

    if (not v.pauseAtUpperBounds or v.pauseAtUpperBounds == {}) then 
        v.pauseAtUpperBounds = ColorCycle_DefaultPauseAtUpperBounds();
    end
    
    if (not v.text) then
        if (v.cycleType == "ToolTipText") then
            ColorCyclePrint("No ToolTipText text passed to ColorCycle_AddTableEntry for UI element "..v.globalName.."!");
            return false;
        else
            v.text = "";
        end
    end

    -- Validate each part
    for member, value in v do
        local validationResult;
        validationResult = ColorCycle_ValidateEntryElement(member, value, v.globalName, v.ID);
        
        if (not validationResult) then
            valid = false;
        end
    end
    
    -- make some adjustment to the parameters to make sure things go ok
    local validated = { };
    if (valid) then
        validated = ColorCycle_TweekEntryElements(v);
    end
    
    return valid, validated;
end
    

    
-----------------------------------------------------------------------------------
-- Check an individual element of a new entry
-- This gets called for each parameter passed to ColorCycle_AddTableEntry()
-- It checks each parameter to make sure it has a default value,
-- is within legal bounds, etc.
-----------------------------------------------------------------------------------
function ColorCycle_ValidateEntryElement(member, value, globalName, theID)

    --ColorCycleDebug("Entering ValidateEntryElement: member = "..member.." value = "..asText(value));

    local UIelementName = "[ID = \""..theID.."\" global = \""..globalName.."\"]";
    
    -- Check the cycleType
    if (member == "cycleType") then
        if (type(value) ~= "string") then 
            ColorCyclePrint("Invalid cycleType sent to ColorCycle_AddTableEntry for UI element "..UIelementName..": should be a string.");
            return false;
        else
            local typeList = "";
            local foundType = false;
            for i = 1, table.getn(COLORCYCLE_CYCLETYPES) do
                typeList = typeList..COLORCYCLE_CYCLETYPES[i];
                if (COLORCYCLE_CYCLETYPES[i] == value) then foundType = true; end
                if (i ~= table.getn(COLORCYCLE_CYCLETYPES)) then typeList = typeList..", "; end
            end
            
            if (not foundType) then
                ColorCyclePrint("Invalid cycleType sent to ColorCycle_AddTableEntry for UI element "..UIelementName..": must be one of "..typeList..".");            
                return false;
            end
        end
    end

    -- Check the cycleSpeed
    if (member == "cycleSpeed") then
        if (type(value) ~= "number") then 
            ColorCyclePrint("Invalid cycleSpeed sent to ColorCycle_AddTableEntry for UI element "..UIelementName..": should be a number.");
            return false;
        end
    end

    -- Check the varySpeedFactor
    if (member == "varySpeedFactor") then
        if (type(value) ~= "number") then 
            ColorCyclePrint("Invalid varySpeedFactor sent to ColorCycle_AddTableEntry for UI element "..UIelementName..": should be a number.");
            return false;
        end
    end

    -- Check the colorStart, colorEnd, lockedValues. okToWait
    if (member == "colorStart" or member == "colorEnd" or member == "lockedValues" or member == "followList" or
        member == "okToWait" or member == "pauseAtLowerBounds" or member == "pauseAtLowerBounds") then
        if (value ~= nil and type(value) ~= "table") then 
            if (type(value) == "string") then 
                ColorCyclePrint(member.." sent to ColorCycle_AddTableEntry is specified as a string, not a table.  Please enclose in {}. [was "..member.." for UI element "..UIelementName..".");
                return false;
            else
                ColorCyclePrint("Invalid "..member.." list sent to ColorCycle_AddTableEntry for UI element "..UIelementName..".");
                return false;
            end
        elseif ((value ~= nil or value ~= {nil, nil, nil} or value ~= {}) and table.getn(value) ~= 3) then
            ColorCyclePrint("Invalid number of elements for "..member.." list sent to ColorCycle_AddTableEntry for UI element "..UIelementName..".  Must be a three element table.");
            return false;       
        else
            for k,v in value do
                if (v ~= nil and 
                        ((member == "colorStart" or member == "colorEnd" or member == "lockedValues" or
                          member == "followList" or member == "pauseAtLowerBounds" or member == "pauseAtLowerBounds" or
                          member == "endCondition") 
                          and (type(v) ~= "number" and type(v) ~= "boolean")) 
                        or
                        (member == "okToWait" and type(v) ~= "boolean")
                   ) then
                    ColorCyclePrint("Invalid "..member.." type sent to ColorCycle_AddTableEntry for UI element "..UIelementName..".");
                    return false;
                else
                    if (member == "followList") then
                        if (v ~= nil and (v < 0 or v > 3)) then
                            ColorCyclePrint("Invalid "..member.." sent to ColorCycle_AddTableEntry for UI element "..UIelementName..".  Must be between 0 and 3.");
                            return false;
                        end
                    elseif (member == "okToWait") then
                        if (v ~= nil and (v ~= true and v ~= false)) then
                            ColorCyclePrint("Invalid "..member.." sent to ColorCycle_AddTableEntry for UI element "..UIelementName..".  Must be true or false.");
                            return false;
                        end
                    elseif (member == "pauseAtLowerBounds" or member == "pauseAtLowerBounds") then
                        if (v ~= nil and (type(v) == "number" and (v < 0 or v > 60)) or (type(v) == "boolean" and (v ~= true and v~= false))) then
                            ColorCyclePrint("Invalid "..member.." sent to ColorCycle_AddTableEntry for UI element "..UIelementName..".  Must be false or between 0 and 60.");
                            return false;
                        end                    
                    else
                        if (v ~= nil and (type(v) == "number" and (v < 0.0 or v > 1.0)) or (type(v) == "boolean" and (v ~= true and v~= false))) then
                            ColorCyclePrint("Invalid "..member.." sent to ColorCycle_AddTableEntry for UI element "..UIelementName..".  Must be false or between 0 and 1.");
                            return false;
                        end
                    end
                end    
            end
        end
    end
    
    -- Check the redBounds, greenBounds, blueBounds
    if (member == "redBounds" or member == "greenBounds" or member == "blueBounds") then
        if (value ~= nil and type(value) ~= "table") then 
            if (type(value) == "string") then 
                ColorCyclePrint(member.." sent to ColorCycle_AddTableEntry is specified as a string, not a table.  Please enclose in {}. [was "..member.." for UI element "..UIelementName..".");
                return false;
            else
                ColorCyclePrint("Invalid "..member.." list sent to ColorCycle_AddTableEntry for UI element "..UIelementName..".");
                return false;
            end
        elseif (value ~= nil and table.getn(value) ~= 2) then
            ColorCyclePrint("Invalid number of elements for "..member.." list sent to ColorCycle_AddTableEntry for UI element "..UIelementName..".  Must be a two element table.");
            return false;       
        else
            for k,v in value do
                if (v ~= nil and type(v) ~= "number") then
                    ColorCyclePrint("Invalid "..member.." sent to ColorCycle_AddTableEntry for UI element "..UIelementName..".");
                    return false;
                elseif (v ~= nil and (v < 0.0 or v > 1.0)) then
                    ColorCyclePrint("Invalid "..member.." sent to ColorCycle_AddTableEntry for UI element "..UIelementName..".  Must be between 0 and 1.");
                    return false;
                end
            end
        end        
    end

    -- Check the text
    if (member == "text") then
        if (type(value) ~= "string") then 
            ColorCyclePrint("Invalid text sent to ColorCycle_AddTableEntry for UI element "..UIelementName..": should be a string.");
            return false;
        end
    end

    -- Check the started flag
    if (member == "started") then
        if (type(value) ~= "boolean") then 
            ColorCyclePrint("Invalid started sent to ColorCycle_AddTableEntry for UI element "..UIelementName..": should be a boolean.");
            return false;
        end
    end

    -- Check the goto
    if (member == "goto") then
        if (value ~= nil and type(value) ~= "string" and type(value) ~= "number") then 
            ColorCyclePrint("Invalid "..member.." sent to ColorCycle_AddTableEntry for UI element "..UIelementName..": should be a string, number, or nil.");
            return false;
        end
    end

    -- Check the startingFunction and endingFunction
    if (member == "startingFunction" or member == "endingFunction") then
        if (value ~= nil and type(value) ~= "table") then 
            ColorCyclePrint("Invalid "..member.." sent to ColorCycle_AddTableEntry for UI element "..UIelementName..": should be a table.");
            return false;
        end
    end

    -- No problems. 
    return true;
end



-----------------------------------------------------------------------------------
-- Tweek the elements (adjust stuff based on other stuff - wow my brain is gone.  just gone.)
-- This just does more dummy checking on parameters sent to ColorCycle_AddTableEntry()
-----------------------------------------------------------------------------------
function ColorCycle_TweekEntryElements(v)

    -- make sure our bounds have default values
    ColorCycleDebug("TweekEntryElements: on entry v.blueBounds[1] = "..asText(v.blueBounds[1]));
    ColorCycleDebug("TweekEntryElements: on entry v.blueBounds[2] = "..asText(v.blueBounds[2]));
    if (v.redBounds[1] == nil) then v.redBounds[1] = 0.0; end
    if (v.redBounds[2] == nil) then v.redBounds[2] = 1.0; end
    if (v.greenBounds[1] == nil) then v.greenBounds[1] = 0.0; end
    if (v.greenBounds[2] == nil) then v.greenBounds[2] = 1.0; end
    if (v.blueBounds[1] == nil) then v.blueBounds[1] = 0.0; end
    if (v.blueBounds[2] == nil) then v.blueBounds[2] = 1.0; end
    
    -- make sure our bounds are in-line with our start and end values
    if (v.colorStart[1] and v.colorEnd[1]) then
        if (v.redBounds[1] < math.min(v.colorStart[1],v.colorEnd[1])) then v.redBounds[1] = math.min(v.colorStart[1],v.colorEnd[1]); end
        if (v.redBounds[2] > math.max(v.colorStart[1],v.colorEnd[1])) then v.redBounds[2] = math.max(v.colorStart[1],v.colorEnd[1]); end
    end
    if (v.colorStart[2] and v.colorEnd[2]) then
        if (v.greenBounds[1] < math.min(v.colorStart[2],v.colorEnd[2])) then v.greenBounds[1] = math.min(v.colorStart[2],v.colorEnd[2]); end
        if (v.greenBounds[2] > math.max(v.colorStart[2],v.colorEnd[2])) then v.greenBounds[2] = math.max(v.colorStart[2],v.colorEnd[2]); end
    end    
    ColorCycleDebug("TweekEntryElements: v.blueBounds[1] = "..v.blueBounds[1]);
    ColorCycleDebug("TweekEntryElements: v.blueBounds[2] = "..v.blueBounds[2]);
    if (v.colorStart[3] and v.colorEnd[3]) then
        if (v.blueBounds[1] < math.min(v.colorStart[3],v.colorEnd[3])) then v.blueBounds[1] = math.min(v.colorStart[3],v.colorEnd[3]); end
        if (v.blueBounds[2] > math.max(v.colorStart[3],v.colorEnd[3])) then v.blueBounds[2] = math.max(v.colorStart[3],v.colorEnd[3]); end
    end
    ColorCycleDebug("After Tweeking...");
    ColorCycleDebug("TweekEntryElements: v.blueBounds[1] = "..v.blueBounds[1]);
    ColorCycleDebug("TweekEntryElements: v.blueBounds[2] = "..v.blueBounds[2]);
    
    -- if we have endvalues we're moving toward, don't allow waiting or speed varying
    for i = 1, 3 do
        if (v.colorEnd[i]) then 
            v.okToWait[i] = false;
            v.varySpeedFactor = 0;
        end
    end
    
    -- change pause nils into zeros
    if (not v.pauseAtUpperBounds or v.pauseAtUpperBounds[1] == nil or v.pauseAtUpperBounds[2] == nil or v.pauseAtUpperBounds[3] == nil) then 
        v.pauseAtUpperBounds = {0, 0, 0};
    end
    if (not v.pauseAtLowerBounds or v.pauseAtLowerBounds[1] == nil or v.pauseAtLowerBounds[2] == nil or v.pauseAtLowerBounds[3] == nil) then 
        v.pauseAtLowerBounds = {0, 0, 0};
    end
    
    return v;
end


-----------------------------------------------------------------------------------
-- find the largest gap between start and endpoints for a channel
-----------------------------------------------------------------------------------
function ColorCycle_FindLargestStartEndGap(colorStart, colorEnd)
    local largestGap = 0.0;
    local gapList = { };
    for i = 1, 3 do
        local gap;
        if (type(colorStart[i]) ~= "number" or type(colorEnd[i]) ~= "number") then
            gap = 0;
        else    
            gap = math.abs(colorStart[i] - colorEnd[i]);
        end
        table.insert(gapList, gap);
        if (gap > largestGap) then largestGap = gap; end
    end

    return largestGap, gapList;
end

    
-----------------------------------------------------------------------------------
-- Process a new entry, inserting it into the Cycle_Table
--
-- Each entry will have the following variables:
--   type:              The type of thing we're colorizing ("ToolTipText", "FontText", 
--                                                          "Texture", "FrameBackground", 
--                                                          "FrameBorder", "StatusBar")
--   global:            The name of the global we're colorizing (as specified in the xml file)
--   text:              If we're colorizing ToolTipText, we need to set the text each time we change the color,
--                      so here's where we specify what text
--   startingFunction:  Each entry can call a function before it starts cycling.  You might want to trigger
--                      some piece of code (like play a sound or something), and here's where you'd do it.
--   endCondition:      This is a table of three floats {redVal, greenVal, blueVal} that tells ColorCycle
--                      when (if ever) to stop cycling this entry and move on to the next one in the queue.
--                      For instance, maybe you want to fade down to black {0, 0, 0} and then move on to
--                      the next entry in the queue.
--   endingFunction:    Each entry can call a function when it finishes cycling (i.e. when the endCondition
--                      is met.  You might want to trigger some piece of code (like play a sound or something), 
--                      and here's where you'd do it.
--   goto:              This tells ColorCycle what to do when it meets the endCondition.  This can take the following
--                      values: 
--                              "end"    - stop cycling this entry entirely.  do not cycle anything else in the queue.
--                              "random" - pick a random entry in the queue and start there
--                              number   - go to the specified entry in the queue
--                              nil      - move on to the next position in the queue, looping back to the first when at the end
--   started:           Gets set to true once this entry starts cycling
--   colorChannel:      This is a list of three elements, one for each of the three color channels: red, green, blue
--                      Entry 1 = red.  Entry 2 = green.  Entry 3 = blue.  These define what each channel is doing.
--
-- Each colorChannel will have the following variables:
--   val:               The current value of this channel - anywhere from 1.0 to 0.0.
--   valOrig:           The initial value of this channel.  We might want a channel to always start at it's initial
--                      value, so we store that here for when/if we ever reset this channel.
--   dir:               The direction this channel is moving.  1 means it's incrementing, -1 means it's decrementing.
--                      0 means it's remaing constant.
--   dirOrig:           The original starting direction we were moving in.  For example, we might want to start at
--                      0.5 and move up to 1.0, always starting in a positive direction.  This variable remembers
--                      our starting direction in case we ever reset this channel.
--   waitFor:           Sometimes we might want a channel to stop changing until another channel reaches a value.
--                      For example, we might want the red channel to remain at full value until the green channel
--                      gets to full value (creating pure yellow).  This flag can be set to a channel number that
--                      we want to wait on (i.e. set it to 2 for the green channel).
--   waitVal:           When we're waiting for another channel to reach a certain value, this speficies the value
--                      we're waiting on.
--   cs:                The amount to change this channel each tick.  Anything from 0.0 to 1.0, although typically
--                      you'd want a value around 0.01 or so.
--   basecs:            This stores the base amount to change.  We have this variable because sometimes we might
--                      modify the current amount we're changing (cs), and then want to revert back to our normal
--                      change amount.
--   csFactor:          Sometimes, we might want a channel to vary how fast its changing.  This represents a factor of
--                      how much to vary the change.  Typically an integer like 2 or 3 to make it 2 or 3 times slower,
--                      or a fraction like 0.5 to make it twice as fast.
--   minVal:            The minimum value this channel is allowed to reach.  We might set this to 0.3 if we want a channel
--                      to never go totally black, or set it to 1.0 to keep a channel constantly full on.
--   maxVal:            The maximum value this channel is allowed to reach.  We might set this to 0.0 if we want a channel
--                      to stay black, or set it to 1.0 to keep a channel constantly full on.
--   followList:        This is a list of integers, like {2, 3} that specifies which (if any) channels this one is
--                      following.  By following, I mean staying the same value.  For example, we might want the red
--                      and green channels to always be the same.  To do that, we could set red to follow green.  Then,
--                      red's value would always be the same as green's.
--   okToWait:          This flag specifies if it's ok for this channel to wait on another.  If we don't want to
--                      allow a channel to ever wait, we'd set this to true.  
--   stayBright:        Set to true if you never want all 3 color channels to ever go to 0
--   pauseUpper:        If we want this channel to pause for x seconds when it reaches its maxVal, we'd set this.
--   pauseLower:        If we want this channel to pause for x seconds when it reaches its minVal, we'd set this.
--   pauseTimeout:      This specifies when to quit pausing.
--   locked:            If we want to lock this channel (not let it change ever), we'd set this
-----------------------------------------------------------------------------------
function ColorCycle_ProcessNewTableEntry(v)
    ColorCycleTron("ColorCycle_ProcessNewTableEntry");    
    
    -- if we have colorEnd (bounds we're moving toward), each channel will need its own speed
    -- in order to reach the colorEnd at the same time
    local cycleSpeedList = { };
    if (v.colorStart == false or v.colorStart[1] == false or v.colorStart[2] == false or v.colorStart[3] == false or
        v.colorEnd == false or v.colorEnd[1] == false or v.colorEnd[2] == false or v.colorEnd[3] == false) then
        cycleSpeedList = {v.cycleSpeed, v.cycleSpeed, v.cycleSpeed};
    else
        -- find the largest gap between start and endpoints for a channel
        local largestGap, gapList = ColorCycle_FindLargestStartEndGap(v.colorStart, v.colorEnd);

        local steps = largestGap / v.cycleSpeed;
        if (steps == 0) then
            cycleSpeedList = {0, 0, 0};
        else
            -- figure out what the speeds will need to be for the other 2 channels to match the first channel
            for i = 1, 3 do
                table.insert(cycleSpeedList, gapList[i]/steps);
            end
        end    
    end
        
    -- store the three channel bounds for easier reference
    local minValues = {v.redBounds[1], v.greenBounds[1], v.blueBounds[1]};
    local maxValues = {v.redBounds[2], v.greenBounds[2], v.blueBounds[2]};
    
    ColorCycleDebug("ProcessNewTableEntry: ("..v.globalName..") minValues = "..asText(minValues));
    ColorCycleDebug("ProcessNewTableEntry: ("..v.globalName..") maxValues = "..asText(maxValues));
    
    -- build the table entry
    tmpTable = { };
    tmpTable.type          = v.cycleType;
    tmpTable.global        = v.globalName;
    tmpTable.text          = v.text;
    
    -- store the startingFunction, endingFunction, goto, and endcondition
    tmpTable.startingFunction = v.startingFunction;       -- a function call
    tmpTable.endCondition     = v.endCondition;           -- {r,g,b}
    tmpTable.endingFunction   = v.endingFunction;         -- a function call
    tmpTable.goto             = v.goto;                   -- "end", "random", a Qpos, or nil for next pos
    tmpTable.started          = false;                    -- false if we haven't started this queue entry yet
    tmpTable.stayBright       = v.stayBright;                   

    tmpTable.colorChannel  = { };
    --ColorCycleDebug("For ["..v.globalName.."]");
    
    -- for each color channel
    for i = 1, 3 do
    
        -- figure out if we're supposed to follow another color
        local colorsToFollow = { };
        if (v.followList[i] ~= 0 and v.followList[i] ~= nil) then
            for j = 1, 3 do
                if (i ~= j) then
                    if (v.followList[j] ~= 0 and v.followList[j] ~= nil) then
                        table.insert(colorsToFollow, j);
                    end
                end
            end
        end
        
        tmpColorChannel = { };
        tmpColorChannel.dir          = 1;
        tmpColorChannel.waitFor      = 0;
        tmpColorChannel.waitVal      = 0;
        tmpColorChannel.basecs       = cycleSpeedList[i];
        tmpColorChannel.cs           = cycleSpeedList[i];
        tmpColorChannel.csFactor     = v.varySpeedFactor;
        tmpColorChannel.minVal       = math.min(math.max(minValues[i], 0.0), 1.0);
        tmpColorChannel.maxVal       = math.max(math.min(maxValues[i], 1.0), 0.0);
        tmpColorChannel.followList   = colorsToFollow;
        tmpColorChannel.okToWait     = v.okToWait[i];
        tmpColorChannel.pauseUpper   = v.pauseAtUpperBounds[i];
        tmpColorChannel.pauseLower   = v.pauseAtLowerBounds[i];
        tmpColorChannel.pauseTimeout = 0;

        -- if we have a specific start/end value, make that our "top" and "bottom" limit
        -- else just use the absolute max and min values of our color range
        if (not v.colorStart[i]) then
            tmpColorChannel.bottomVal = tmpColorChannel.minVal;
        else
            tmpColorChannel.bottomVal = v.colorStart[i];
        end
        if (not v.colorEnd[i]) then
            tmpColorChannel.topVal = tmpColorChannel.maxVal;
        else
            tmpColorChannel.topVal = v.colorEnd[i];
        end
        
        -- determine our starting value: either we're locked to another color
        tmpColorChannel.locked       = v.lockedValues[i];
        if (tmpColorChannel.locked) then
            tmpColorChannel.val = tmpColorChannel.locked;
            ColorCycleDebug("ProcessNewTableEntry: ("..v.globalName..") channel "..i.." LOCKED to a color.");
        -- we specified a starting value
        elseif (v.colorStart[i] ~= false) then
            --tmpColorChannel.val = math.min(math.max(colorStart[i], tmpColorChannel.minVal), tmpColorChannel.maxVal);
            tmpColorChannel.val = v.colorStart[i];
            ColorCycleDebug("ProcessNewTableEntry: ("..v.globalName..") channel "..i.." SPECIFIED a color.");
            
        -- or we pick a random one
        else        
            local colorValRange = {minValues[i], maxValues[i]};        
            tmpColorChannel.val = colorValRange[math.random(2)];      
            ColorCycleDebug("ProcessNewTableEntry: ("..v.globalName..") channel "..i.." RANDOM color.");
        end

        ColorCycleDebug("ProcessNewTableEntry: ("..v.globalName..") channel "..i.." minVal = "..asText(tmpColorChannel.minVal));
        ColorCycleDebug("ProcessNewTableEntry: ("..v.globalName..") channel "..i.." maxVal = "..asText(tmpColorChannel.maxVal));
        ColorCycleDebug("ProcessNewTableEntry: ("..v.globalName..") channel "..i.." start value = "..asText(tmpColorChannel.val));

        -- remember our original values so we can reset ourself if needed
        tmpColorChannel.dirOrig      = tmpColorChannel.dir;
        tmpColorChannel.valOrig      = tmpColorChannel.val;

        table.insert(tmpTable.colorChannel, tmpColorChannel);
    end
    
    -- finally!  insert this entry into the Cycle_Table
    ColorCycle_AddEntry(tmpTable, v.ID);
    ColorCycleDebug("Added to Cycle_Table: "..v.globalName);
    
end



-----------------------------------------------------------------------------------
-- Add an entry to the list of objects to color cycle
-----------------------------------------------------------------------------------
function ColorCycle_AddEntry(theEntry, theID)

    -- see if this ID is already in the table
    local IDindex = 0;
    for i = 1, table.getn(Cycle_Table) do
        if (Cycle_Table[i].ID == theID) then
            IDindex = i;
            break;
        end
    end
    
    -- if there's already an entry for this ID, add the new entry to the queue
    if (IDindex ~= 0) then
         table.insert(Cycle_Table[IDindex].Q, theEntry);
    
    -- else create a new queue for this ID and add the entry
    else
        table.insert(Cycle_Table, ColorCycle_CreateNewEntry(theID));
        table.insert(Cycle_Table[table.getn(Cycle_Table)].Q, theEntry);
    end
end


-----------------------------------------------------------------------------------
-- Create an entry to the list of objects to color cycle
-----------------------------------------------------------------------------------
function ColorCycle_CreateNewEntry(theID)
    tmpTable      = { };
    tmpTable.ID   = theID;
    tmpTable.Q    = { };
    tmpTable.Qpos = 1;
    
    return tmpTable;
end


-----------------------------------------------------------------------------------
-- Remove an entry from the list of objects to color cycle
-----------------------------------------------------------------------------------
function ColorCycle_RemoveEntry(entryID)
    result = false;
    
    for i = 1, table.getn(Cycle_Table) do
        if (Cycle_Table[i].ID == entryID) then
            table.remove(Cycle_Table, i);
            ColorCycleDebug("Removed from Cycle_Table: "..entryID);
            result = true;
            break;
        end
    end
    
    return result;
end


-----------------------------------------------------------------------------------
-- return true if theGlobal is already in the cycle_table
-----------------------------------------------------------------------------------
function ColorCycle_InCycleTable(theID)
    local result = false;
    for i = 1, table.getn(Cycle_Table) do
        if (Cycle_Table[i].ID == theID) then
            result = true;
            break;
        end
    end
    
    return result;
end


-----------------------------------------------------------------------------------
-- Update
-----------------------------------------------------------------------------------
function ColorCycle_OnUpdate()
    if (ColorCycle_Config.Enabled) then
        ColorCycle_CycleColors();
    end
end


-----------------------------------------------------------------------------------
-- For each entry, cycle its colors
-----------------------------------------------------------------------------------
function ColorCycle_CycleColors()

    local i = 1;
    
    -- for each entry in the cycle_table
    while (i <= table.getn(Cycle_Table)) do
    
        -- check to see if we should perform the start action
        if (not Cycle_Table[i].Q[Cycle_Table[i].Qpos].started) then
            Cycle_Table[i].Q[Cycle_Table[i].Qpos].started = true;
            
            -- call our starting function
            if (Cycle_Table[i].Q[Cycle_Table[i].Qpos].startingFunction) then
                ColorCycle_CallFunction(unpack(Cycle_Table[i].Q[Cycle_Table[i].Qpos].startingFunction));
            end    
        end
        
        -- cycle the entry
        ColorCycle_CycleAnEntry(i);

        -- if we have an end condition
        if (Cycle_Table[i].Q[Cycle_Table[i].Qpos].endCondition) then
        
            -- figure out our current color
            local currentColor = { };
            for j = 1, 3 do
                table.insert(currentColor, Cycle_Table[i].Q[Cycle_Table[i].Qpos].colorChannel[j].val);
            end
            
            -- if we've reached an end condition
            if (tostring(currentColor[1]) == tostring(Cycle_Table[i].Q[Cycle_Table[i].Qpos].endCondition[1]) and
                tostring(currentColor[2]) == tostring(Cycle_Table[i].Q[Cycle_Table[i].Qpos].endCondition[2]) and
                tostring(currentColor[3]) == tostring(Cycle_Table[i].Q[Cycle_Table[i].Qpos].endCondition[3])) then

                -- perform the ending function call for this queue entry
                if (Cycle_Table[i].Q[Cycle_Table[i].Qpos].endingFunction) then
                    ColorCycle_CallFunction(unpack(Cycle_Table[i].Q[Cycle_Table[i].Qpos].endingFunction));
                end
                
                -- reset some stuff for this entry
                ColorCycle_ResetColorEntry(i);

                -- determine the next position to move to in the queue
                -- end this queue
                local nextQpos = nil;
                if (tostring(Cycle_Table[i].Q[Cycle_Table[i].Qpos].goto) == "end") then
                    nextQpos = nil;

                -- pick a random position
                elseif (tostring(Cycle_Table[i].Q[Cycle_Table[i].Qpos].goto) == "random") then
                    nextQpos = math.random(table.getn(Cycle_Table[i].Q));

                -- go to the specified position in the queue
                elseif (type(Cycle_Table[i].Q[Cycle_Table[i].Qpos].goto) == "number") then
                    if (Cycle_Table[i].Q[Cycle_Table[i].Qpos].goto < 1) then Cycle_Table[i].Q[Cycle_Table[i].Qpos].goto = 1; end
                    if (Cycle_Table[i].Q[Cycle_Table[i].Qpos].goto > table.getn(Cycle_Table[i].Q)) then Cycle_Table[i].Q[Cycle_Table[i].Qpos].goto = table.getn(Cycle_Table[i].Q); end
                    nextQpos = Cycle_Table[i].Q[Cycle_Table[i].Qpos].goto;

                -- if no position is specified, move to the next position, looping back around to the first position if needed
                else
                    if (Cycle_Table[i].Qpos == table.getn(Cycle_Table[i].Q)) then
                        nextQpos = 1;
                    else
                        nextQpos = Cycle_Table[i].Qpos + 1;
                    end
                end

                -- if we're ending this queue          
                if (not nextQpos) then
                    ColorCycle_RemoveEntry(Cycle_Table[i].ID);
                    i = i - 1;
                else
                    -- reset the started flag for this entry
                    Cycle_Table[i].Q[Cycle_Table[i].Qpos].started = false;
                    Cycle_Table[i].Qpos = nextQpos;            
                end                 
            end
        end
        
        -- advance to the next entry
        i = i + 1;        
    end
end


-----------------------------------------------------------------------------------
-- Update the colors for an entry
-----------------------------------------------------------------------------------
function ColorCycle_CycleAnEntry(i)

    -- only color cycle stuff that's visible
    if (getglobal(Cycle_Table[i].Q[Cycle_Table[i].Qpos].global) and getglobal(Cycle_Table[i].Q[Cycle_Table[i].Qpos].global):IsVisible()) then

        -- update the color values
        Cycle_Table[i].Q[Cycle_Table[i].Qpos].colorChannel = ColorCycle_UpdateColor(Cycle_Table[i].Q[Cycle_Table[i].Qpos].colorChannel, Cycle_Table[i].Q[Cycle_Table[i].Qpos].stayBright);

        -- less typing = fewer mistakes
        local red       = Cycle_Table[i].Q[Cycle_Table[i].Qpos].colorChannel[1].val;
        local green     = Cycle_Table[i].Q[Cycle_Table[i].Qpos].colorChannel[2].val;
        local blue      = Cycle_Table[i].Q[Cycle_Table[i].Qpos].colorChannel[3].val;
        local theText   = Cycle_Table[i].Q[Cycle_Table[i].Qpos].text;
        local theGlobal = Cycle_Table[i].Q[Cycle_Table[i].Qpos].global;
        local theType   = Cycle_Table[i].Q[Cycle_Table[i].Qpos].type;
        
        if (type(theGlobal) == "string") then
            globalVal = getglobal(theGlobal);
        else
            globalVal = theGlobal;
        end
        
        ColorCycle_SetColorValueForUIElement(globalVal, theType, red, blue, green, theText)
        
    end
end


-----------------------------------------------------------------------------------
-- Set the colors for an entry - the specific calls for each type of UI element
-----------------------------------------------------------------------------------
function ColorCycle_SetColorValueForUIElement(globalVal, theType, red, blue, green, theText)
    -- set the onscreen colors based on the type of the global
    if (theType == "ToolTipText") then
        globalVal:SetText(theText, red, blue, green);
    elseif (theType == "FontText") then
        globalVal:SetTextColor(red, blue, green);
    elseif (theType == "Texture") then
        globalVal:SetVertexColor(red, blue, green);
    elseif (theType == "FrameBackground") then
        globalVal:SetBackdropColor(red, blue, green);
    elseif (theType == "FrameBorder") then
        globalVal:SetBackdropBorderColor(red, blue, green);
    elseif (theType == "StatusBar") then
        globalVal:SetStatusBarColor(red, blue, green);
    end  
end


-----------------------------------------------------------------------------------
-- Calculate the R, G, and B values for an entry
--
-- ** The main logic of the AddOn **
-----------------------------------------------------------------------------------
function ColorCycle_UpdateColor(colorChannel, stayBright)
    
    -- for each channel (r, g, b)
    for i = 1, 3 do
    
        -- if this color channel is locked, don't change its value
        if (not colorChannel[i].locked) then
        
            -- check to see if we should just follow another color
            if (i ~= 1 and table.getn(colorChannel[i].followList) > 0) then
            
                -- if we're blue, see if we should follow red
                if (i == 2) then
                    local foundRed = false;
                    for j = 1, table.getn(colorChannel[i].followList) do
                        if (colorChannel[i].followList[j] == 1) then
                            foundRed = true;
                            break;
                        end
                    end
                    if (foundRed) then
                        colorChannel[i].val = colorChannel[1].val;
                    end

                -- if we're green, see if we should follow red or blue
                elseif (i == 3) then
                    local foundRedorGreen = 0;
                    for j = 1, table.getn(colorChannel[i].followList) do
                        if (colorChannel[i].followList[j] == 1 or colorChannel[i].followList[j] == 2) then
                            foundRedorGreen = colorChannel[i].followList[j];
                            break;
                        end
                    end
                    if (foundRedorGreen ~= 0) then
                        colorChannel[i].val = colorChannel[foundRedorGreen].val;
                    end
                end
            else
            
                -- if we've reached our top limit or our bottom limit
                local reachedTop = false;
                local reachedBottom = false;
                if (colorChannel[i].val >= colorChannel[i].maxVal) then reachedTop = true; end
                if (colorChannel[i].val <= colorChannel[i].minVal) then reachedBottom = true; end
                if (reachedTop or reachedBottom) then

                    -- if we need to pause at this limit
                    if (((reachedTop and colorChannel[i].pauseUpper ~= 0) or (reachedBottom and colorChannel[i].pauseLower ~= 0)) and
                         colorChannel[i].pauseTimeout == 0) then
                        if (reachedTop) then
                            colorChannel[i].pauseTimeout = GetTime() + colorChannel[i].pauseUpper;
                        else
                            colorChannel[i].pauseTimeout = GetTime() + colorChannel[i].pauseLower;
                        end
                    else

                        -- if we're done pausing at a limit boundary
                        if (GetTime() >= colorChannel[i].pauseTimeout) then
                
                            colorChannel[i].pauseTimeout = 0;

                            -- if we're waiting for another color to cycle
                            if (colorChannel[i].waitFor ~= 0) then

                                -- see if we should keep waiting
                                if (tostring(colorChannel[colorChannel[i].waitFor].val) == tostring(colorChannel[i].waitVal)) then
                                    colorChannel[i].waitFor = 0;
                                end
                            end

                            -- decide a new direction to move in
                            if (colorChannel[i].waitFor == 0) then
                                local rn = math.random(4);

                                -- either we cycle back in the other direction
                                if (rn <= 2 or not colorChannel[i].okToWait) then
                                    if (colorChannel[i].val >= colorChannel[i].maxVal) then
                                        colorChannel[i].dir = -1;

                                        -- don't ever let all three channels cycle toward 0 at once
                                        -- Hmmm -- this might be a problem?
                                        if (stayBright) then
                                            if (colorChannel[1].dir == -1 and colorChannel[2].dir == -1 and colorChannel[3].dir == -1) then
                                                colorChannel[i].dir = 1;
                                            end
                                        end    
                                    else
                                        colorChannel[i].dir = 1;
                                    end
                                    colorChannel[i].waitFor = 0;

                                    -- chance to cycle regular speed, fast, or superfast
                                    colorChannel[i].cs = ColorCycle_GetCycleSpeed(colorChannel[i].basecs, colorChannel[i].csFactor);

                                -- or we wait where we are for another color to cycle
                                else

                                    -- make sure one of the other channels isn't already waiting
                                    local okToWait = false;
                                    local waitList = { };
                                    local otherChannels = {1, 2, 3};
                                    for n = 1, 3 do
                                        if (otherChannels[n] == i) then
                                            table.remove(otherChannels, n);
                                            break;
                                        end
                                    end     

                                    -- if another channel isn't already waiting AND we're not following another color
                                    if (colorChannel[otherChannels[1]].waitFor == 0 and colorChannel[otherChannels[2]].waitFor == 0 and
                                            table.getn(colorChannel[i].followList) == 0) then
                                        okToWait = true;
                                        waitList = otherChannels;
                                    end

                                    -- make sure we're not waiting on a locked color channel that'll never move
                                    local m = 1;
                                    while (m <= table.getn(waitList)) do
                                        if (colorChannel[waitList[m]].locked ~= nil) then
                                            table.remove(waitList, m);
                                        else
                                            m = m + 1;
                                        end                            
                                    end

                                    -- if no other color is waiting and there are still possible colors to wait on
                                    -- then we can wait
                                    if (okToWait and table.getn(waitList) > 0) then
                                        colorChannel[i].waitFor  = waitList[math.random(table.getn(waitList))];

                                        -- either wait for the minimum, middle, or max value of the channel
                                        local minWait = colorChannel[colorChannel[i].waitFor].minVal;
                                        local maxWait = colorChannel[colorChannel[i].waitFor].maxVal;
                                        waitDest = {minWait, (minWait+maxWait)/2.0, maxWait};
                                        colorChannel[i].waitVal = waitDest[math.random(3)];
                                        colorChannel[i].dir = 0;

                                    -- else don't wait and just cycle
                                    else
                                        if (colorChannel[i].val == 1.0) then
                                            colorChannel[i].dir = -1;
                                            
                                            if (stayBright) then
                                                if (colorChannel[1].dir == -1 and colorChannel[2].dir == -1 and colorChannel[3].dir == -1) then
                                                    colorChannel[i].dir = 1;
                                                end
                                            end    
                                        else
                                            colorChannel[i].dir = 1;
                                        end      
                                        colorChannel[i].waitFor = 0;

                                        -- chance to cycle regular speed, fast, or superfast
                                        colorChannel[i].cs = ColorCycle_GetCycleSpeed(colorChannel[i].basecs, colorChannel[i].csFactor);
                                    end
                                end
                            end    
                        end    
                    end    -- if we need to pause at this limit
                end

                -- adjust the value for this color channel
                colorChannel[i].val = colorChannel[i].val + (colorChannel[i].cs * colorChannel[i].dir);

                -- bounds check
                if (colorChannel[i].val > colorChannel[i].maxVal) then colorChannel[i].val = colorChannel[i].maxVal; end
                if (colorChannel[i].val < colorChannel[i].minVal) then colorChannel[i].val = colorChannel[i].minVal; end
            end    
        end
    end
    
    --ColorCycleDebug("r = "..colorChannel[1].val.."  g = "..colorChannel[2].val.."  b = "..colorChannel[3].val);
        
    return colorChannel;    
end


-----------------------------------------------------------------------------------
-- Return a cycle speed given a base cycle speed (basecs)
-----------------------------------------------------------------------------------
function ColorCycle_GetCycleSpeed(basecs, factor)
    local rn2 = math.random(10);
    if (rn2 <= 6 or factor == 0) then
         result = basecs;
    elseif (rn2 <= 8) then
         result = basecs * (2 * factor);
    elseif (rn2 <= 9) then
         result = basecs * (4 * factor);
    else
         result = basecs * (0.5 * factor);    
    end   
    
    return result;
end


-----------------------------------------------------------------------------------
-- figure out the speed needed to move from color1 to color2 such that it'd be
-- the same amount of time as moving from black to white
-----------------------------------------------------------------------------------
function ConstantSpeed(theSpeed, colorStart, colorEnd)
    local steps = 1.0 / theSpeed;
    local largestGap = ColorCycle_FindLargestStartEndGap(colorStart, colorEnd);
    local baseSpeed = largestGap / steps;
    
    return baseSpeed;
end


-----------------------------------------------------------------------------------
-- Reset some stuff for an entry
-----------------------------------------------------------------------------------
function ColorCycle_ResetColorEntry(i)
    for channel = 1, 3 do
        Cycle_Table[i].Q[Cycle_Table[i].Qpos].colorChannel[channel].dir = Cycle_Table[i].Q[Cycle_Table[i].Qpos].colorChannel[channel].dirOrig;
        Cycle_Table[i].Q[Cycle_Table[i].Qpos].colorChannel[channel].val = Cycle_Table[i].Q[Cycle_Table[i].Qpos].colorChannel[channel].valOrig;
        Cycle_Table[i].Q[Cycle_Table[i].Qpos].colorChannel[channel].waitFor = 0;
    end
end

--============================================================================================--
--============================================================================================--
--                                                                                            --
--                              UTILITY FUNCTIONS                                             --
--                                                                                            --
--============================================================================================--
--============================================================================================--


-----------------------------------------------------------------------------------
-- Returns true if ColorCycle is installed
-----------------------------------------------------------------------------------
function ColorCycle_Exists()
    return true;
end
    
    
-----------------------------------------------------------------------------------
-- Call a function with its arguments
-----------------------------------------------------------------------------------
function ColorCycle_CallFunction(func,...)
    if (arg) then
        func(unpack(arg));
    else
        func();
    end
end


-----------------------------------------------------------------------------------
-- Toggle the AddOn on/off
-----------------------------------------------------------------------------------
function ColorCycle_Toggle(toggle)
    if (toggle == nil) then 
        ColorCycle_Config.Enabled = not ColorCycle_Config.Enabled;
    else
        if (toggle == 1) then
            ColorCycle_Config.Enabled = true;
        else
            ColorCycle_Config.Enabled = false;
        end
    end
end


-----------------------------------------------------------------------------------
-- Our own print function
-----------------------------------------------------------------------------------
function ColorCyclePrint(theMsg, r, g, b)
    
    -- 0.73, 0.53, 0.87
    if (r == nil) then r = 1.00; end
    if (g == nil) then g = 0.50; end
    if (b == nil) then b = 0.50; end

    if (type(theMsg) == "table") then
        Print(asText(theMsg), r, g, b);
    else
        Print(theMsg, r, g, b);
    end
end


-----------------------------------------------------------------------------------
-- Show the help text
-----------------------------------------------------------------------------------
function ColorCycle_ShowHelp()

    DEFAULT_CHAT_FRAME:AddMessage(" ");
    ColorCycle_AddHelpMessage(COLORCYCLE_COLOR_CLOSE..COLORCYCLE_COLOR_GREEN.."ColorCycle, an AddOn by Cragganmore.  "..COLORCYCLE_COLOR_CLOSE..COLORCYCLE_COLOR_BLUE.."Version "..COLORCYCLE_VERSION, nil, nil, nil);
    DEFAULT_CHAT_FRAME:AddMessage(COLORCYCLE_COLOR_GREEN.."Usage: "..COLORCYCLE_COLOR_CLOSE..COLORCYCLE_COLOR_BLUE.."/colorcycle <command>"..COLORCYCLE_COLOR_CLOSE..COLORCYCLE_COLOR_GREEN.." or "..COLORCYCLE_COLOR_CLOSE..COLORCYCLE_COLOR_BLUE.."/cc <command>"..COLORCYCLE_COLOR_CLOSE);
    ColorCycle_AddHelpMessage("  "..COLORCYCLE_ON_CMD.."|"..COLORCYCLE_OFF_CMD, "Turns ColorCycle on or off.", nil, ColorCycle_Config.Enabled);

    ColorCycle_AddHelpMessage("  "..COLORCYCLE_HELP_CMD, "Displays this message.", nil, nil);
    DEFAULT_CHAT_FRAME:AddMessage(" ");
    ColorCycle_AddHelpMessage("Example: /colorcycle on", nil, nil, nil);
end


-----------------------------------------------------------------------------------
-- Prints a formatted help message
-----------------------------------------------------------------------------------
function ColorCycle_AddHelpMessage(command, detail, status, enabled)

    local message = "";
    
    if (detail == nil) then
        message = COLORCYCLE_COLOR_BLUE..command..COLORCYCLE_COLOR_CLOSE;
    else
        message = COLORCYCLE_COLOR_BLUE..command..COLORCYCLE_COLOR_CLOSE..COLORCYCLE_COLOR_GREEN..": "..detail..COLORCYCLE_COLOR_CLOSE;
    end
    
    if (enabled == nil) then
        DEFAULT_CHAT_FRAME:AddMessage(message);
        return;
    end
    
    if (status == "" or status == nil) then
        if (enabled and enabled ~= 0) then
            DEFAULT_CHAT_FRAME:AddMessage(message..COLORCYCLE_COLOR_BLUE.." (enabled)"..COLORCYCLE_COLOR_CLOSE);
        else
            DEFAULT_CHAT_FRAME:AddMessage(message..COLORCYCLE_COLOR_GREY.." (disabled)"..COLORCYCLE_COLOR_CLOSE);
        end
    else
        if (enabled and enabled ~= 0) then
            DEFAULT_CHAT_FRAME:AddMessage(message..COLORCYCLE_COLOR_BLUE.." (set to "..status..")"..COLORCYCLE_COLOR_CLOSE);
        else
            DEFAULT_CHAT_FRAME:AddMessage(message..COLORCYCLE_COLOR_CLOSE.." (set to "..status..")"..COLORCYCLE_COLOR_CLOSE);
        end     
    end
end


-----------------------------------------------------------------------------------
-- Debug function to view any lua object as text
-----------------------------------------------------------------------------------
function asText(obj)

    visitRef = {}
    visitRef.n = 0

    asTxRecur = function(obj, asIndex)
        if type(obj) == "table" then
            if visitRef[obj] then
                return "@"..visitRef[obj]
            end
            visitRef.n = visitRef.n +1
            visitRef[obj] = visitRef.n

            local begBrac, endBrac
            if asIndex then
                begBrac, endBrac = "[{", "}]"
            else
                begBrac, endBrac = "{", "}"
            end
            local t = begBrac
            local k, v = nil, nil
            repeat
                k, v = next(obj, k)
                if k ~= nil then
                    if t > begBrac then
                        t = t..", "
                    end
                    t = t..asTxRecur(k, 1).."="..asTxRecur(v)
                end
            until k == nil
            return t..endBrac
        else
            if asIndex then
                -- we're on the left side of an "="
                if type(obj) == "string" then
                    return obj
                else
                    return "["..obj.."]"
                end
            else
                -- we're on the right side of an "="
                if type(obj) == "string" then
                    return '"'..obj..'"'
                else
                    return tostring(obj)
                end
            end
        end
    end -- asTxRecur

    return asTxRecur(obj)
end -- asText


--============================================================================================--
--============================================================================================--
--                                                                                            --
--                              DEBUG FUNCTIONS                                               --
--                                                                                            --
--============================================================================================--
--============================================================================================--


-----------------------------------------------------------------------------------
-- Toggle debugging on/off
-----------------------------------------------------------------------------------
function ColorCycle_ToggleDebug(toggle)
    if (toggle == nil) then 
        COLORCYCLE_DEBUG = not COLORCYCLE_DEBUG;
    else
        if (toggle == 1) then
            COLORCYCLE_DEBUG = true;
            Print("ColorCycle: Debug On", 1.0, 1.0, 0.5);
        else
            COLORCYCLE_DEBUG = false;
            Print("ColorCycle: Debug Off", 1.0, 1.0, 0.5);
        end
    end

end

-----------------------------------------------------------------------------------
-- Show debug text
-----------------------------------------------------------------------------------
function ColorCycleDebug(theMsg, r, g, b, level)
    if (not r) then r = 1.0; end
    if (not g) then g = 0.5; end
    if (not b) then b = 0.5; end
    if (not level) then level = -1; end

    if ((COLORCYCLE_DEBUG and level < 0) or (COLORCYCLE_DEBUG and COLORCYCLE_DEBUG_LEVEL >= level)) then
        DEFAULT_CHAT_FRAME:AddMessage(theMsg, r, g, b);
    end
end


-----------------------------------------------------------------------------------
-- Print a trace message
-----------------------------------------------------------------------------------
function ColorCycleTron(theMsg)
    if (COLORCYCLE_DEBUG or COLORCYCLE_TRON) then
        DEFAULT_CHAT_FRAME:AddMessage(theMsg, 0.3, 0.6, 0.3);
    end
end


-----------------------------------------------------------------------------------
-- Show a debug report
-----------------------------------------------------------------------------------
function ColorCycle_Report(full, entry)
    local startIndex = 1;
    local endIndex = table.getn(Cycle_Table);
    
    if (entry) then
        startIndex = tonumber(entry);
        endIndex = startIndex;
    end
    
    ColorCyclePrint("Cycle_Table ("..table.getn(Cycle_Table).." entries):");
    for i = startIndex, endIndex do
        ColorCyclePrint("Entry "..i.." = "..Cycle_Table[i].ID);
        if (full) then
            ColorCyclePrint("  Queue has "..table.getn(Cycle_Table[i].Q).." element(s):");
            for j = 1, table.getn(Cycle_Table[i].Q) do
                ColorCyclePrint(asText(Cycle_Table[i].Q[j]));
                ColorCyclePrint("-----------------------");
            end
            ColorCyclePrint("========================");
        else
            ColorCyclePrint("  Queue has "..table.getn(Cycle_Table[i].Q).." element(s)");
        end
    end
end


-----------------------------------------------------------------------------------
-- Show a debug report
-----------------------------------------------------------------------------------
function ColorCycle_QList()
    ColorCyclePrint("Cycle_Table ("..table.getn(Cycle_Table).." entries):");
    for i = 1, table.getn(Cycle_Table) do
        ColorCyclePrint("Entry "..i.." = "..Cycle_Table[i].ID);
        ColorCyclePrint("  Queue has "..table.getn(Cycle_Table[i].Q).." element(s):");
        for j = 1, table.getn(Cycle_Table[i].Q) do
            ColorCyclePrint("     "..Cycle_Table[i].Q[j].global);
            ColorCyclePrint("-----------------------");
        end
        ColorCyclePrint("========================");
    end
end



-----------------------------------------------------------------------------------
-- Functions to return copies of default tables
-- so that we don't have to make copies manually
-----------------------------------------------------------------------------------
function ColorCycle_DefaultStartValues()
    local tmp = { };
    for i = 1, table.getn(COLORCYCLE_DEFAULT_STARTVALUES) do
        table.insert(tmp, COLORCYCLE_DEFAULT_STARTVALUES[i]);
    end
    
    return tmp;
end
function ColorCycle_DefaultEndValues()
    local tmp = { };
    for i = 1, table.getn(COLORCYCLE_DEFAULT_ENDVALUES) do
        table.insert(tmp, COLORCYCLE_DEFAULT_ENDVALUES[i]);
    end
    
    return tmp;
end
function ColorCycle_DefaultLockedValues()
    local tmp = { };
    for i = 1, table.getn(COLORCYCLE_DEFAULT_LOCKEDVALUES) do
        table.insert(tmp, COLORCYCLE_DEFAULT_LOCKEDVALUES[i]);
    end
    
    return tmp;
end
function ColorCycle_DefaultRedBounds()
    local tmp = { };
    for i = 1, table.getn(COLORCYCLE_DEFAULT_REDBOUNDS) do
        table.insert(tmp, COLORCYCLE_DEFAULT_REDBOUNDS[i]);
    end
    
    return tmp;
end
function ColorCycle_DefaultGreenBounds()
    local tmp = { };
    for i = 1, table.getn(COLORCYCLE_DEFAULT_GREENBOUNDS) do
        table.insert(tmp, COLORCYCLE_DEFAULT_GREENBOUNDS[i]);
    end
    
    return tmp;
end
function ColorCycle_DefaultBlueBounds()
    local tmp = { };
    for i = 1, table.getn(COLORCYCLE_DEFAULT_BLUEBOUNDS) do
        table.insert(tmp, COLORCYCLE_DEFAULT_BLUEBOUNDS[i]);
    end
    
    return tmp;
end
function ColorCycle_DefaultFollowList()
    local tmp = { };
    for i = 1, table.getn(COLORCYCLE_DEFAULT_FOLLOWLIST) do
        table.insert(tmp, COLORCYCLE_DEFAULT_FOLLOWLIST[i]);
    end
    
    return tmp;
end
function ColorCycle_DefaultOkToWait()
    local tmp = { };
    for i = 1, table.getn(COLORCYCLE_DEFAULT_OKTOWAIT) do
        table.insert(tmp, COLORCYCLE_DEFAULT_OKTOWAIT[i]);
    end
    
    return tmp;
end
function ColorCycle_DefaultPauseAtLowerBounds()
    local tmp = { };
    for i = 1, table.getn(COLORCYCLE_DEFAULT_PAUSEATLOWERBOUNDS) do
        table.insert(tmp, COLORCYCLE_DEFAULT_PAUSEATLOWERBOUNDS[i]);
    end
    
    return tmp;
end
function ColorCycle_DefaultPauseAtUpperBounds()
    local tmp = { };
    for i = 1, table.getn(COLORCYCLE_DEFAULT_PAUSEATUPPERBOUNDS) do
        table.insert(tmp, COLORCYCLE_DEFAULT_PAUSEATUPPERBOUNDS[i]);
    end
    
    return tmp;
end
function ColorCycle_CopyTable(theTable)
    local tmp = { };
    for i = 1, table.getn(theTable) do
        table.insert(tmp, theTable[i]);
    end
    
    return tmp;
end
