--============================================================================================--
--============================================================================================--
--                                                                                            --
--                              API FUNCTIONS                                                --
--                                                                                            --
--============================================================================================--
--============================================================================================--




-----------------------------------------------------------------------------------
--  Remove an entry from the Cycle_Table
-----------------------------------------------------------------------------------
function ColorCycle_Remove(theID)
    if (ColorCycle_InCycleTable(theID)) then 
        ColorCycle_RemoveEntry(theID);
    end    
end


function ColorCycle_FadeInElements(...)
    
end

--[[            
            -- cause the hotkey buttons to cycle through fading up and down
            -- after each key fades it, it calls PlaySound("TellMessage")
            local buttonColor1 = {1.0, 1.0, 1.0};
            local buttonColor2 = {1.0, 0.0, 0.0};
            for i = 0, 7 do
                getglobal("ActionButton"..tostring(i).."Icon"):SetVertexColor(0,0,0);
                ColorCycle_AddTableEntry( 
                    {
                        ID              = "ActionButtons";
                        cycleType       = "Texture";
                        globalName      = "TitanPanelBackground"..tostring(i);
                        colorStart      = buttonColor1;
                        colorEnd        = buttonColor2;
                        okToWait        = {false, false, false};
                        cycleSpeed      = 0.1;
                        endCondition    = buttonColor2;
                        unique          = false;
                    }
                );
            end
            for i = 7, 0, -1 do
                ColorCycle_AddTableEntry( 
                    {
                        ID              = "ActionButtons";
                        cycleType       = "Texture";
                        globalName      = "TitanPanelBackground"..tostring(i);
                        colorStart      = buttonColor2;
                        colorEnd        = buttonColor1;
                        okToWait        = {false, false, false};
                        cycleSpeed      = 0.1;
                        endCondition    = buttonColor1;
                        unique          = false;
                    }
                );
            end
]]--

-----------------------------------------------------------------------------------
-- make a UI element repeatedly change from one color to another color and back
--
-- ColorCycle_AlternateColors()
--      ID          = string;
--      globalName  = nameOfUIelement
--      cycleType   = "ToolTipText" | "FontText" | "Texture" | "FrameBackground" | "FrameBorder" | "StatusBar"
--      colorStart  = { red, green, blue }
--      colorEnd    = { red, green, blue }
--      speed       = amountToChangePerTick (0.05 is an average value)
--      text        = string (only needed when doing a "ToolTipText" cycle)
-----------------------------------------------------------------------------------
function ColorCycle_AlternateColors(...)

    for i = 1, table.getn(arg) do      
        local v = arg[i];
        
        if (v.color) then v.colorStart = v.color; end
        if (v.color1) then v.colorStart = v.color1; end
        if (v.color2) then v.colorEnd = v.color2; end
        
        -- check parameters and set defaults
        if (v.remove) then
            if (ColorCycle_InCycleTable(v.ID)) then ColorCycle_RemoveEntry(v.ID); end
            return;
        end  
        if (not v.colorStart) then
            ColorCyclePrint("Invalid parameters to ColorCycle_AlternateColors.  Missing colorStart.");
            return;
        end
        if (type(v.colorStart) == "string") then
            if (COLORCYCLE_BASECOLORS[v.colorStart]) then
                v.colorStart = ColorCycle_CopyTable(COLORCYCLE_BASECOLORS[v.colorStart]);
            else
                ColorCyclePrint("Invalid parameters to ColorCycle_AlternateColors.  Unknown value for colorStart ("..v.colorStart..")");
                return;
            end
        end
        if (not v.colorEnd) then
            ColorCyclePrint("Invalid parameters to ColorCycle_AlternateColors.  Missing colorEnd.");
            return;
        end
        if (type(v.colorEnd) == "string") then
            if (COLORCYCLE_BASECOLORS[v.colorEnd]) then
                v.colorEnd = ColorCycle_CopyTable(COLORCYCLE_BASECOLORS[v.colorEnd]);
            else
                ColorCyclePrint("Invalid parameters to ColorCycle_AlternateColors.  Unknown value for colorEnd ("..v.colorEnd..")");
                return;
            end
        end
        if (type(v.endCondition) == "string") then
            if (COLORCYCLE_BASECOLORS[v.endCondition]) then
                v.endCondition = ColorCycle_CopyTable(COLORCYCLE_BASECOLORS[v.endCondition]);
            else
                ColorCyclePrint("Invalid parameters to ColorCycle_AlternateColors.  Unknown value for endCondition ("..v.endCondition..")");
                return;
            end
        end
        if (not v.text) then v.theText = ""; end
        if (not v.speed) then v.speed  = COLORCYCLE_DEFAULT_GLOWSPEED; end

        -- add this to the cycle table
        ColorCycle_AddTableEntry(
            {
                ID                  = v.ID;
                cycleType           = v.cycleType;
                globalName          = v.globalName;
                cycleSpeed          = ConstantSpeed(v.speed, v.colorStart, v.colorEnd);
                colorStart          = v.colorStart;
                colorEnd            = v.colorEnd;
                okToWait            = {false, false, false};
                text                = v.text;
                unique              = v.unique;
            }
        );    
    end    
end


-----------------------------------------------------------------------------------
-- make a UI element repeatedly pulse (slowly) from its color to white
--
-- ColorCycle_PulseWhite()
--      ID          = string;
--      globalName  = nameOfUIelement
--      cycleType   = "ToolTipText" | "FontText" | "Texture" | "FrameBackground" | "FrameBorder" | "StatusBar"
--      color       = { red, green, blue }
--      pause       = seconds to pause when at colorStart    
--      speed       = amountToChangePerTick (0.05 is an average value)
--      text        = string (only needed when doing a "ToolTipText" cycle)
-----------------------------------------------------------------------------------
function ColorCycle_PulseWhite(...)

    for i = 1, table.getn(arg) do      
        local v = arg[i];
        
        if (v.color) then v.colorStart = v.color; end
        if (v.color1) then v.colorStart = v.color1; end
        if (v.color2) then v.colorEnd = v.color2; end

        -- check parameters and set defaults
        if (v.remove) then
            if (ColorCycle_InCycleTable(v.ID)) then ColorCycle_RemoveEntry(v.ID); end
            return;
        end  
        if (not v.pause) then v.pause  = COLORCYCLE_DEFAULT_PAUSE; end
        if (not v.text) then v.theText = ""; end
        if (not v.speed) then v.speed  = COLORCYCLE_DEFAULT_GLOWSPEED; end

        -- call ColorCycle_PulseColor with the 2nd color = white
        ColorCycle_PulseColor(
            {
                ID                  = v.ID;
                globalName          = v.globalName;
                cycleType           = v.cycleType;
                colorStart          = v.colorStart;
                colorEnd            = {1, 1, 1};
                pause               = v.pause;
                speed               = v.speed;
                text                = v.text;
                remove              = v.remove;
                startingFunction    = v.startingFunction;
                endingFunction      = v.endingFunction;
                endCondition        = v.endCondition;
                goto                = v.goto;
                unique              = v.unique;
            }
        );    
    end    
end


-----------------------------------------------------------------------------------
-- make a UI element repeatedly pulse (slowly) from its color to white
--
-- ColorCycle_PulseColor()
--      ID          = string;
--      globalName  = nameOfUIelement
--      cycleType   = "ToolTipText" | "FontText" | "Texture" | "FrameBackground" | "FrameBorder" | "StatusBar"
--      colorStart      = { red, green, blue }
--      colorEnd      = { red, green, blue }
--      pause       = seconds to pause when at colorStart    
--      speed       = amountToChangePerTick (0.05 is an average value)
--      text        = string (only needed when doing a "ToolTipText" cycle)
-----------------------------------------------------------------------------------
function ColorCycle_PulseColor(...)

    for i = 1, table.getn(arg) do      
        local v = arg[i];
        
        if (v.color) then v.colorStart = v.color; end
        if (v.color1) then v.colorStart = v.color1; end
        if (v.color2) then v.colorEnd = v.color2; end

        -- check parameters and set defaults
        if (v.remove) then
            if (ColorCycle_InCycleTable(v.ID)) then ColorCycle_RemoveEntry(v.ID); end
            return;
        end  
        if (not v.colorStart) then
            ColorCyclePrint("Invalid parameters to ColorCycle_PulseColor.  Missing colorStart.");
            return;
        end
        if (type(v.colorStart) == "string") then
            if (COLORCYCLE_BASECOLORS[v.colorStart]) then
                v.colorStart = ColorCycle_CopyTable(COLORCYCLE_BASECOLORS[v.colorStart]);
            else
                ColorCyclePrint("Invalid parameters to ColorCycle_PulseColor.  Unknown value for colorStart ("..v.colorStart..")");
                return;
            end
        end
        if (not v.colorEnd) then
            ColorCyclePrint("Invalid parameters to ColorCycle_PulseColor.  Missing colorEnd.");
            return;
        end
        if (type(v.colorEnd) == "string") then
            if (COLORCYCLE_BASECOLORS[v.colorEnd]) then
                v.colorEnd = ColorCycle_CopyTable(COLORCYCLE_BASECOLORS[v.colorEnd]);
            else
                ColorCyclePrint("Invalid parameters to ColorCycle_PulseColor.  Unknown value for colorEnd ("..v.colorEnd..")");
                return;
            end
        end
        if (type(v.endCondition) == "string") then
            if (COLORCYCLE_BASECOLORS[v.endCondition]) then
                v.endCondition = ColorCycle_CopyTable(COLORCYCLE_BASECOLORS[v.endCondition]);
            else
                ColorCyclePrint("Invalid parameters to ColorCycle_PulseColor.  Unknown value for endCondition ("..v.endCondition..")");
                return;
            end
        end
        if (not v.pause) then v.pause  = COLORCYCLE_DEFAULT_PAUSE; end
        if (not v.text) then v.theText = ""; end
        if (not v.speed) then v.speed  = COLORCYCLE_DEFAULT_GLOWSPEED; end

        -- add this to the cycle table
        ColorCycle_AddTableEntry(
            {
                ID                  = v.ID;
                cycleType           = v.cycleType;
                globalName          = v.globalName;
                cycleSpeed          = ConstantSpeed(v.speed, v.colorStart, v.colorEnd);
                colorStart          = v.colorStart;
                colorEnd            = v.colorEnd;
                okToWait            = {false, false, false};
                pauseAtUpperBounds  = {v.pause, v.pause, v.pause};
                text                = v.text;
                startingFunction    = v.startingFunction;
                endingFunction      = v.endingFunction;
                endCondition        = v.endCondition;
                goto                = v.goto;
                unique              = v.unique;
            }
        );    
        
    end    
end


-----------------------------------------------------------------------------------
-- make a UI element repeatedly flash (quickly) from its color to white
-- ColorCycle_FlashWhite()
--      ID          = string;
--      globalName  = nameOfUIelement
--      cycleType   = "ToolTipText" | "FontText" | "Texture" | "FrameBackground" | "FrameBorder" | "StatusBar"
--      color       = { red, green, blue }
--      pause       = seconds to pause when at color    
--      speed       = amountToChangePerTick (0.05 is an average value)
--      text        = string (only needed when doing a "ToolTipText" cycle)
-----------------------------------------------------------------------------------
function ColorCycle_FlashWhite(...)

    for i = 1, table.getn(arg) do      
        local v = arg[i];
        
        if (v.color) then v.colorStart = v.color; end
        if (v.color1) then v.colorStart = v.color1; end
        if (v.color2) then v.colorEnd = v.color2; end

        -- check parameters and set defaults
        if (v.remove) then
            if (ColorCycle_InCycleTable(v.ID)) then ColorCycle_RemoveEntry(v.ID); end
            return;
        end    
        if (not v.text) then v.theText = ""; end

        -- add this to the cycle table
        ColorCycle_FlashColor(
            {
                ID                  = v.ID;
                cycleType           = v.cycleType;
                globalName          = v.globalName;
                cycleSpeed          = 0.07;
                colorStart          = v.colorStart;
                colorEnd            = {1, 1, 1};
                text                = v.text;
                startingFunction    = v.startingFunction;
                endingFunction      = v.endingFunction;
                endCondition        = v.endCondition;
                goto                = v.goto;
                unique              = v.unique;
            }
        );  
    end    
end


-----------------------------------------------------------------------------------
-- make a UI element repeatedly flash (quickly) from its color to white
-- ColorCycle_FlashColor()
--      ID          = string;
--      globalName  = nameOfUIelement
--      cycleType   = "ToolTipText" | "FontText" | "Texture" | "FrameBackground" | "FrameBorder" | "StatusBar"
--      colorStart      = { red, green, blue }
--      colorEnd      = { red, green, blue }
--      pause       = seconds to pause when at colorStart    
--      speed       = amountToChangePerTick (0.05 is an average value)
--      text        = string (only needed when doing a "ToolTipText" cycle)
-----------------------------------------------------------------------------------
function ColorCycle_FlashColor(...)

    for i = 1, table.getn(arg) do      
        local v = arg[i];
        
        if (v.color) then v.colorStart = v.color; end
        if (v.color1) then v.colorStart = v.color1; end
        if (v.color2) then v.colorEnd = v.color2; end

        -- check parameters and set defaults
        if (v.remove) then
            if (ColorCycle_InCycleTable(v.ID)) then ColorCycle_RemoveEntry(v.ID); end
            return;
        end    
        if (not v.colorStart) then
            ColorCyclePrint("Invalid parameters to ColorCycle_FlashColor.  Missing colorStart.");
            return;
        end
        if (type(v.colorStart) == "string") then
            if (COLORCYCLE_BASECOLORS[v.colorStart]) then
                v.colorStart = ColorCycle_CopyTable(COLORCYCLE_BASECOLORS[v.colorStart]);
            else
                ColorCyclePrint("Invalid parameters to ColorCycle_FlashColor.  Unknown value for colorStart ("..v.colorStart..")");
                return;
            end
        end
        if (not v.colorEnd) then
            ColorCyclePrint("Invalid parameters to ColorCycle_FlashColor.  Missing colorEnd.");
            return;
        end
        if (type(v.colorEnd) == "string") then
            if (COLORCYCLE_BASECOLORS[v.colorEnd]) then
                v.colorEnd = ColorCycle_CopyTable(COLORCYCLE_BASECOLORS[v.colorEnd]);
            else
                ColorCyclePrint("Invalid parameters to ColorCycle_FlashColor.  Unknown value for colorEnd ("..v.colorEnd..")");
                return;
            end
        end
        if (not v.text) then v.theText = ""; end

        -- add this to the cycle table
        ColorCycle_AddTableEntry(
            {
                ID                  = v.ID;
                cycleType           = v.cycleType;
                globalName          = v.globalName;
                cycleSpeed          = ConstantSpeed(0.05, v.colorStart, v.colorEnd);
                colorStart          = v.colorStart;
                colorEnd            = v.colorEnd;
                okToWait            = {false, false, false};
                text                = v.text;
                startingFunction    = v.startingFunction;
                endingFunction      = v.endingFunction;
                endCondition        = v.endCondition;
                goto                = v.goto;
                unique              = v.unique;
           }
        );    
    end    
end
