
local function defaults() return {
	SAY = { active = 'on' },
	PARTY = { active = 'on' },
	GUILD = { active = 'on' },
	OFFICER = { active = 'on' },
	YELL = { active = 'off' },
	WHISPER = { active = 'on' },
	REPLY = { active = 'on' },
	EMOTE = { active = 'on' },
	TEXT_EMOTE = { active = 'on' },

	MONSTER_EMOTE = { active = 'on' },
	MONSTER_SAY = { active = 'on' },
	MONSTER_WHISPER = { active = 'on' },
	MONSTER_YELL = { active = 'on' },

	SYSTEM = { active = 'on', filter = 'auction|online|offline|outbid' },

	CHANNEL1 = { active = 'off' },
	CHANNEL2 = { active = 'off' },
	CHANNEL3 = { active = 'off' },
	CHANNEL4 = { active = 'off' },
}; end

Echo_Watched2 = defaults();

Echo_SoundFile = 'blip2.wav';
Echo_Enabled = true;

local savedOnEvent;
local lastTime = GetTime();
local recentMessages = {}

local function print(msg, color)
	color = color or { r = 1, g = 1, b = 1 };
	DEFAULT_CHAT_FRAME:AddMessage('Echo: ' .. msg, color.r, color.g, color.b);
end

function Echo_OnLoad()
	SlashCmdList['ECHO'] = Echo_SlashHandler;
	SLASH_ECHO1 = '/echo';

	savedOnEvent = ChatFrame_OnEvent;
	ChatFrame_OnEvent = Echo_OnEventHook;
end

function Echo_OnEventHook(event)
	local msgtype = strsub(event, 10);
	local info = ChatTypeInfo[msgtype];
	if event == "CHAT_MSG_CHANNEL" then
		msgtype = "CHANNEL" .. tostring(arg8);
		info = ChatTypeInfo[msgtype];
	end

	local text = arg1;
	local sender = arg2 or '';

	local watch = Echo_Watched2[msgtype] or { active = 'off' };
	local fromself = sender == UnitName("player");

	-- check if message should be echoed
	if Echo_Enabled and (watch.active == 'all' or (watch.active == 'self' and fromself)
		or (watch.active == 'on' and not fromself)) then

		-- format text
		if sender == '' then
			-- no sender
			-- do nothing
		elseif msgtype == 'MONSTER_EMOTE' or msgtype == 'EMOTE' then
			-- emote text without embedded name
			text = sender .. ' ' .. text;
		elseif msgtype == 'SYSTEM' or msgtype == 'SKILL' or msgtype == 'LOOT'
			or strsub(msgtype,1,7) == 'COMBAT_'
			or strsub(msgtype,1,6) == 'SPELL_'
			or msgtype == 'TEXT_EMOTE' then
			-- message text with embedded name or without name
			-- do nothing
		else
			-- somebody said something...
			text = sender .. ': ' .. text;
		end

		local show = false;
		if watch.filter == nil then
			show = true;
		else
			local f = string.lower(watch.filter);
			local ltext = string.lower(text);
			local sub;
			for sub in string.gfind(f, '[^|]+') do
				if string.find(ltext, sub) then
					show = true;
					break;
				end
			end
		end

		if show then
			-- show message
			local id = tostring(msgtype)..'#'..tostring(sender)..'#'..tostring(text);
	                if Echo_TryFloatMsg(text, info, id) then
				-- play sound
				PlaySoundFile(Echo_SoundFile);
			end
		end
	end

	-- call original OnEvent function
	savedOnEvent(event);
end

function Echo_TryFloatMsg(msg, color, id)
	-- check for duplicate messages within the last second
	if GetTime() - lastTime > 1.0 then
		recentMessages = {};
	end
	local shown = false;
	if recentMessages[id] == nil then
		recentMessages[id] = true;
		Echo_FloatMsg(msg, color);
		shown = true;
	end
	lastTime = GetTime();
	return shown;
end

function Echo_FloatMsg(msg, color)
	UIErrorsFrame:AddMessage(msg, color.r, color.g, color.b, 1.0, UIERRORS_HOLD_TIME);
end

local function formatCurrentValue(msgtype)
	local t = Echo_Watched2[msgtype] or { active = 'off' };
	if t.filter then
		return t.active .. ', ' .. t.filter;
	else
		return t.active;
	end
end

function Echo_SlashHandler(command)
	local i,j, a1, a2, a3 = string.find(command, '^([^ ]+)[ ]*([^ ]*)[ ]*(.*)$');
	a2 = a2 or '';
	a3 = a3 or '';
	if a1 == nil then
		-- /echo
		print('Usage: /echo <type> <which> [filter]');
		print('Which: on, off, self, all');
		print('Use "/echo list" or "/echo listall" to see a list of targets.');
		print('Use "/echo list <name>" to list targets containing <name>.');
		print('Use "/echo sound <filename>" to set the audio file.');
		print('Use "/echo on" or "/echo off" to enable / disable echo.');
		if not Echo_Enabled then
			print('Echo is disabled.', { r = 1, g = 0, b = 0 });
		end
	elseif a1 == 'on' then
		print('Enabled.');
		Echo_Enabled = true;
	elseif a1 == 'off' then
		print('Disabled.');
		Echo_Enabled = false;
	elseif a1 == 'list' and a2 ~= '' then
		-- /echo list <pattern>
		a2 = string.upper(a2);
		print('Message types containing "' .. a2 .. '":');
		local unknown = false;
		table.foreach(ChatTypeInfo,
			function(msgtype, info)
				if string.find(msgtype, a2) then
					print(msgtype .. ' (' .. formatCurrentValue(msgtype) .. ')', info);
				end
			end
		);
	elseif a1 == 'list' then
		-- /echo list
		print('Message types:');
		table.foreach(Echo_Watched2,
			function(msgtype, filter)
				print(msgtype .. ' (' .. formatCurrentValue(msgtype) .. ')', ChatTypeInfo[msgtype]);
			end
		);
	elseif a1 == 'listall' then
		-- /echo listall
		print('All message types:');
		table.foreach(ChatTypeInfo,
			function(msgtype, info)
				print(msgtype .. ' (' .. formatCurrentValue(msgtype) .. ')', info);
			end
		);
	elseif a1 == 'sound' then
		if a2 == '' then
			-- /echo sound
			print('Current sound file: ' .. Echo_SoundFile);
		else
			-- /echo sound <file>
			Echo_SoundFile = a2;
			print('Sound file set to: ' .. Echo_SoundFile);
		end
	elseif a2 == '' then
		-- /echo <type>
		local msgtype = string.upper(a1);
		if ChatTypeInfo[msgtype] then
			print(msgtype .. ' (' .. formatCurrentValue(msgtype) .. ')', ChatTypeInfo[msgtype]);
		else
			print('"' .. msgtype .. '" is not a valid message type');
		end
	else
		-- /echo <type> <filter>
		local msgtype = string.upper(a1);
		if ChatTypeInfo[msgtype] then
			if a2 == 'on' or a2 == 'off' or a2 == 'self' or a2 == 'all' then
				Echo_Watched2[msgtype] = { active = a2 };
				if a3 ~= '' then
					Echo_Watched2[msgtype].filter = a3;
				end
				print(msgtype .. ' = ' .. formatCurrentValue(msgtype), ChatTypeInfo[msgtype]);
			else
				print('"' .. a2 .. '" is not a valid filter');
				print('Use "on", "off", "self" or "all".');
			end
		else
			print('"' .. msgtype .. '" is not a valid message type');
		end
	end
end
